#!/usr/bin/env python3
"""
versions.py

Outputs a list of software versions used that night. Rewritten by Dylan Gatlin
 based on spVersion by Elena Malanushenko
"""

import subprocess as sub
import tpmdata
import signal
import multiprocessing

__version__ = '3.1.0'


def tpm_timeout(sig, frame):
    raise TimeoutError("TPM Timeout")


def get_tpm_packet(out_dict):
    tpmdata.tinit()
    data = tpmdata.packet(1, 1)
    for key, val in data.items():
        out_dict[key] = val
    return 0


def main():
    
    softwares, versions = [], []

    sdss_obstools = sub.run('pip list | grep sdss-obstools', shell=True,
                              stdout=sub.PIPE
                           ).stdout.decode('utf-8').strip('\n')
    if len(sdss_obstools) != 0:
        softwares.append(sdss_obstools.split()[0])
        versions.append(sdss_obstools.split()[-1])
    else:
        softwares.append("sdss-obstools")
        versions.append("FAILED")
        
    data = multiprocessing.Manager().dict()
    tpm_thread = multiprocessing.Process(target=get_tpm_packet, args=(data,))
    tpm_thread.start()

    module_help = sub.run("module --help", shell=True, stdout=sub.PIPE,
                              stderr=sub.PIPE
                              ).stderr.decode("utf-8")
    has_module = "command not found" not in module_help
    
    tpm_thread.join(1)
    
    if tpm_thread.is_alive():
        tpm_thread.kill()   
    softwares.append("TPM")
    softwares.append("MCP")
    softwares.append("PLC")
    softwares.append("Fiducials")
    if tpm_thread.is_alive():
        tpm_thread.kill()
        versions.append("FAILED")
        versions.append("FAILED")
        versions.append("FAILED")
        versions.append("FAILED")
    else:
        versions.append(data["tpm_vers"])
        versions.append(data["mcp_vers"])
        versions.append(data["plc_vers"])
        versions.append(data["fid_vers"])
        
    
    if has_module:
        names = ["idlspec2d", "Kronos", "roboscheduler"]
        cmds = ["module load idlspec2d; idlspec2d_version",
                "module load kronos; kronosversion.py",
                "module load roboscheduler; roboschedulerversion.py"]
        for n, c in zip(names, cmds):
            ver = sub.run(c, shell=True, stdout=sub.PIPE, stderr=sub.PIPE
                          ).stdout.decode("utf-8").rstrip('\n')
            softwares.append(n)
            if ver:
                versions.append(ver)
            else:
                versions.append("FAILED")
    
    print('{:-^42}'.format('Other Versions'))
    for s, v in zip(softwares, versions):
        print('{:<20}: {:<20}'.format(s, v))
    print()

    disk_usage = sub.run('df -h | grep -e " /home\|data"', shell=True,
                         stdout=sub.PIPE).stdout.decode("utf-8")
    print(disk_usage)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Plot and analyze new fiducial data before applying it to the mcp.

Originally converted nearly line-by-line by RHL from iop/etc/fiducials.tcl,
hence the generally unpythonic and cryptic syntax.
By John Parejko. 

History of changes
2020-04-02 -Fixed bug to read data file as numpy array, same as it was done for
 gz file. 
           -Reconfigured  how to read path for data file, default from current,
            and 
            optional from args,  no mjd format for directory name. Default dir
             option is ''. 
            
~/sdss4/fiducials/git/trunk/bin/plotMcpFiducials.py -rot -mjd 58799 -t0 1573542725 -t1 1573545137 -reset   -noplot
/Users/elenam/sdss4/fiducials/mcptpm/58799

Program: 
~/sdss4/fiducials/git/trunk/bin/plotMcpFiducials.py 
comm='~/sdss4/fiducials/git/trunk/bin/plotMcpFiducials.py'
eval $comm  <args>

Args: 
-mjd 58799 -az -t0 1573538393 -t1 1573542412  -table az.dat
-mjd 58799 -alt -t0 1573545352 -t1 1573546087  -table alt.dat
-mjd 58799 -rot -t0 1573542725 -t1 1573545137  -table rot.dat  -setCanon 80

-canon -reset -scale -verbose -noplot  -table stdout
            
-scale  az, alt, rot
warning
/Users/elenam/sdss4/fiducials/git/trunk/bin/plotMcpFiducials.py:448:
 RuntimeWarning: invalid value encountered in less
  tmp = fiducials[np.abs(fiducials_deg - 360 - fiducials_deg[i]) < 1]


plotMcpFiducials.py -mjd 58799 -dir data/58799 
-alt -t0 1573545352 -t1 1573546087 -fiducialFile=data/v1_141/alt.dat  
-table stdout -canon -reset -scale -noplot

  
"""
# from sdss.utilities import yanny
from pydl.pydlutils.yanny import yanny
import os
import pwd
import re
import sys
import gzip
import argparse
try:
    import matplotlib.pyplot as pyplot
    pyplot.ion()
    pyplot.rcParams.update({'font.size': 15,
                            'axes.labelsize': 13,
                            'legend.fontsize': 16,
                            'xtick.labelsize': 10,
                            'ytick.labelsize': 10,
                            'axes.linewidth': 2})
    from matplotlib import gridspec
except ImportError:
    pyplot = None

from pathlib import Path
from astropy.time import Time

from sdssobstools import sdss_paths

import numpy as np
np.seterr(all='raise')
np.seterr(invalid='warn')


def getMJD():
    """Get the MJD for the current day as an int."""
    return int(Time.now().mjd)


axisAbbrevs = dict(azimuth="az", altitude="alt", rotator="rot")


def reset_fiducials(fididx, ud_delta, velocity, pos, verbose, axisName, resetFid):
    """
    Reset the position vector <pos> every time it crosses the first
    fiducial in the vector fididx
    """

    # # Find the indices where the velocity changes sign
    # NOTE: we do want to catch the first and last elements!
    sign = np.greater_equal(velocity, 0)
    sign_changes = (np.roll(sign, 1) - sign) != 0
    # add in the last element if we counted the first
    # (so we're counting both) front and back of the array)
    #if sign_changes[0]: sign_changes[-1] = True
    sign_changes[0] = True  # EM
    sign_changes[-1] = True  # EM
    sign_changes = np.arange(0, len(pos))[sign_changes]

    # Generate a vector with the `sweep number', incremented every time
    # the velocity changes sign
    sweep = []                    # which sweep the fiducial belongs to
    for i in range(len(sign_changes) - 1):
        sweep += (sign_changes[i+1] - sign_changes[i])*[i]
    # make sweep and fididx have same length
    sweep += (len(fididx) - len(sweep))*[i]
    sweep = np.array(sweep)

    # print "len(sweep)=", len(sweep),  "len(fididx)=", len(fididx)
    # print "%s    %s     %s   %s " % ("fididx",  "pos1",  "velocity", "sweep ")
    # for i in range(len(fididx)):
    #    print "%s  %4i  %10i  %8i  %s " % ( i, fididx[i], pos[i], velocity[i], sweep[i])

    # Choose the average fiducial as a standard (it should be in the middle of
    # the range), resetting the offset to zero whenever we see it
    nff, ff = 0, 0.0
    for i in range(len(sign_changes)):
        sign_changes = fididx[np.where(sweep == i)]
        sign_changes[np.where(sign_changes > ud_delta)] -= ud_delta

        ff += sum(sign_changes)
        nff += len(sign_changes)

    ff = int(ff/nff + 0.5)
    ffval = 0.0
    # ff=65    # from  RL code  and JP, but only 2 sweeps have this fid  for 58799, my testing data
    # ff=79   #  7 fid availble,  good for reset.
    # Find the indices where we see the ff fiducial,  tmp  <type 'numpy.ndarray'>
    if verbose:
        print("Default resetting fiducial %s,  requested %s" % (ff, resetFid))

    if resetFid is None:
        pass
    else:
        ff = resetFid

    tmp = np.arange(len(pos))[np.logical_or(
        fididx == ff, fididx == ff + ud_delta)]
    if verbose:
        print("Resetting at fiducial %d" % ff)
    if len(tmp) == 0:
        print("ERROR: reseting fiducial does not have values, exit",
              file=sys.stderr)
        sys.exit(1)

    # optimize reset fiducial for rotator - EM test code, not approved to use
    # if axisName=="rotator":
    #    maxind=len(tmp); ff2=ff
    #    for ind in range(1,10):
    #        ff1=ff+ind
    #        tmp = np.arange(len(pos))[np.logical_or(fididx == ff1,
    #  fididx == ff1 + ud_delta)]
    #        if len(tmp)> maxind: maxind=len(tmp); ff2=ff1
    #        ff1=ff-ind
    #        tmp = np.arange(len(pos))[np.logical_or(fididx == ff1,
    #  fididx == ff1 + ud_delta)]
    #        if len(tmp)> maxind: maxind=len(tmp); ff2=ff1
    #    ff=ff2
    #    tmp = np.arange(len(pos))[np.logical_or(fididx == ff,
    #  fididx == ff + ud_delta)]
    #    if verbose:
    #        print "Optimized resetting at fiducial %d" % ff
    #        print "Optimized indices of reset fiducials, ff=%d,
    #   %s " %  (ff,tmp), " (%s)" % len(tmp)

    # Generate a vector of corrections; the loop count is the index
    # of resetting fiducials in tmp array, one per sweep if exist
    corrections = {}   # <type 'dict'>
    for i in range(len(tmp)):
        corrections[sweep[tmp[i]]] = ffval - pos[tmp[i]]

# ---  JP corrections algorithm with EM adapted-----
#  For rot,   6 sweeps,  selected middle fid #60,  but only 2 sweeps
#  have this fiducial.  For each sweep we have been check if reset fid exist,
#  if yes, ok.
# if not, check on sweep below if exist, and next check for sweep up, if exist.
# if both exist, take average with prorated diff.
# if not low, take upper. If not upper, take low.
# the result looks not good.

    # Did any sweeps miss the `reset-canonical' fiducials?
    nsweep = sweep[-1] + 1
    # print "Cycle on sweeps:"

    for s in range(nsweep):
        # print "sweep=", s, " corrections.has_key(s):", corrections.has_key(s)
        if s in corrections:
            continue    # already got it

        # for s0 in range(s, -1, -1):   # JP
        for s0 in range(s-1, -2, -1):   # EM range(start, stop, step)
            # print "    testing low sweep (s0) =", s0, " has correction? ",
            #  corrections.has_key(s0)
            if s0 in corrections:
                break

        # for s1 in range(s, nsweep):  # JP
        for s1 in range(s+1, nsweep+1):  # EM
            # print "    testing upper sweep (s1) =", s1, " has correction? ",
            #  corrections.has_key(s1)
            if s1 in corrections:
                break

        #assert (s0 >= 0) or (s1 < nsweep)
        # JP corrections
        if s0 >= 0:
            if s1 < nsweep:              # we have two points
                #import ipdb
                # ipdb.set_trace()
                # print "    set between s0 and s1"
                corrections[s] = corrections[s0] + \
                    float(s - s0)/(s1 - s0)*(corrections[s1] - corrections[s0])
                # KeyError: 0
            else:
                # print "    set same as s0"
                corrections[s] = corrections[s0]
        else:
            # print "    set same as s1"
            corrections[s] = corrections[s1]

    # Build a corrections vector the same length as e.g. $pos
    corr = np.empty_like(pos) + 1.1e10
    for s, c in list(corrections.items()):
        corr[sweep == s] = c

    # Apply those corrections
    pos += np.int32(corr)

    return pos


def make_fiducials_vector(fididx, extend=True):
    """Make a vector from 0 to number of fiducials that we've seen."""
    if extend:
        return np.arange(0, max(fididx)+1)
    else:
        return np.array(sorted(set(fididx)), dtype=int)

# def read_fiducials(fiducial_file, axisName, readHeader=True):


def read_fiducials(fiducial_file, readHeader=True):
    """
    Read the positions of the fiducials for AXIS into VECS findex, p1 and p2

    FIDUCIAL_FILE may be a filename, or a format expecting a single string
    argument (az, alt, or rot), or an MCP version.
    """
    fiducial_file = Path(fiducial_file)
    if not fiducial_file.exists():
        raise RuntimeError(
            "I can find fiducial file  {}".format(fiducial_file))

    try:
        ffd = fiducial_file.open('r')
    except IOError as e:
        raise RuntimeError("I cannot read %s: %s" % (fiducial_file, e))

    # Read header
    header = {}
    while True:
        line = ffd.readline()

        if not line or re.search(r"^\# Fiducial", line):
            break

        if not readHeader:
            continue

        mat = re.search(r"^#\s*([^:]+):\s*(.*)", line)
        if mat:
            var, val = mat.group(1), mat.group(2)
            if var == "$Name":
                var = "Name"
                val = re.search(r"^([^ ]+)", val).group(1)

            if var == "Canonical fiducial":
                val = int(val)
            elif var == "Scales":
                val = [float(x) for x in val.split()]

            header[var] = val

    # Done with header; read data
    vecNames = [("findex", 0, int), ("pos1", 1, float), ("pos2", 5, float)]
    vecs = {}
    for v, col, t in vecNames:
        vecs[v] = []

    while True:
        fields = ffd.readline().split()

        if not fields:
            break

        for v, col, tt in vecNames:
            vecs[v].append(fields[col])

    ffd.close()

    # Convert to numpy arrays
    for v, col, tt in vecNames:
        vecs[v] = np.array(vecs[v], dtype=tt)

    return fiducial_file, vecs, header


def get_input_file(filename, mjd, mcp_log_dir, verbose=False):
    """Figure out where the file is, and return the full path.
    EM: 2020-08-09 reconfigured get_fiducial_file(), if no arg -dir, 
    look file in current directory,  if -dir <dirname>, read in <dirname>. """

    mcp_log_dir = Path(mcp_log_dir)

    if not filename:
        filename = Path("mcpFiducials-{:.0f}.dat".format(mjd))

    # searching input file in
    # -dir (1),  (.), default - /data/logs/mcp/ starting from 58722,
    # old data ?
    result = mcp_log_dir / filename
    if not result.exists():
        result = result.with_suffix(result.suffix + ".gz")
    if not result.exists():
        print("did not find data or gz file {},  exit".format(result),
              file=sys.stderr)
        sys.exit(1)

    return result.absolute()


def set_canonical_fiducials(axisName, vecs, setCanonical):
    """Set up the canonical fiducial values."""

    if setCanonical:
        temp = setCanonical.split(':')
        ff = int(temp[0])
        if len(temp) == 1:
            ffval = (vecs["true1"][vecs["fididx"] == ff][0])
        else:
            ffval = int(temp[1])
        ss = "(setCanonical:%s:%s)" % (ff, ffval)  # setCanonical
    else:
        if axisName == 'azimuth':
            ff = 19
        elif axisName == 'altitude':
            ff = 1
        elif axisName == 'rotator':
            ff = 80
        ffval = (vecs["true1"][vecs["fididx"] == ff][0])
        #ffval=(vecs["true1"][vecs["fididx"] == ff][0]).copy()
        ss = "(default)"

    return ff, ffval, ss


def plot_one(fig, ax, x, y, xlabel, ylabel, axisName, velocity, ffile,
             ms, t0, tbase):

    xmin = min(x)  # xmax = max(x)
    # ymin = min(y); ymax = max(y)

    if xlabel == "time":
        if t0 != 0:
            xmin = t0 - tbase
        # if args.t1 != 0:
        #     xmax = t1 - tbase

    # set limits to just beyond the data on all sides
    ax.margins(0.1)
    ax.yaxis.grid(True, which='major')

    l = velocity >= 0
    ax.plot(x[l], y[l], "rx", ms=3, mew=3, label='v>=0')   # ms - sym size;
    l = velocity < 0
    ax.plot(x[l], y[l], "g+", ms=3, mew=3, label='v<0')

    if xlabel == 'time' and ms:
        plot_ms(fig, ms['on'], ms['off'], axisName, tbase)

    if xlabel == "time" and tbase:
        # print "Initial time on plot:",tbase-xmin
        ax.set_xlabel("time - %s" % tbase)
    else:
        ax.set_xlabel(xlabel)

    fig.show()


def plot_data(axisName, vecs, index, plotFile, ffile,
              ms, reset, updown, error, t0, tbase):
    """Generate plots of (time,deg,index,etc.) vs. the fiducial positions."""

    def make_filename(axisName, xlabel, ylabel):
        if plotFile:
            filename = "%s-%s-%s" % (axisName, xlabel, ylabel)
            if reset:
                filename += "-R"
            if updown:
                filename += "-U"
            if updown:
                filename += "-E"
            filename += ".png"
        else:
            filename = ''
        return filename

    # Have to delete this from the local mapping, as it breaks matplotlib's
    # TK backend. It will still be set for the shell when we exit.
    if os.environ.get('TCL_LIBRARY', None) is not None:
        del os.environ['TCL_LIBRARY']

    title = axisName
    if reset:
        title += " -reset"
    if updown:
        title += " -updown"
    if error:
        title += " -error"
    title += "\n  Data: %s \n" % ffile

    labels = ['time', 'deg', 'index', 'fididx']
    for ylabel in ['pos1', 'pos2']:
        fig = pyplot.figure(figsize=(13, 8))
        gs = gridspec.GridSpec(2, 2, wspace=0.02)  # , height_ratios=[3, 1])
        ax = []
        #pyplot.rc('figure', titlesize=10)
        pyplot.suptitle(title)
        for i, xlabel in enumerate(labels):
            # right-side plots share the y axis.
            if i % 2 == 0:
                ax.append(pyplot.subplot(gs[i]))
                ax[i].set_ylabel(ylabel)
            else:
                ax.append(pyplot.subplot(gs[i], sharey=ax[i-1]))
                pyplot.setp(ax[i].get_yticklabels(), visible=False)
                ax[i].yaxis.tick_right()

            plotfile = make_filename(axisName, xlabel, ylabel)
            if xlabel == 'index':
                plot_one(fig, ax[i], index, vecs[ylabel], xlabel, ylabel,
                         axisName, vecs['velocity'], ffile, ms, t0, tbase)
            else:
                plot_one(fig, ax[i], vecs[xlabel], vecs[ylabel], xlabel,
                         ylabel, axisName, vecs['velocity'], ffile, ms, t0,
                         tbase)
            # use this for the label, incase we plotted with 'ms'
            if xlabel == 'time':
                ax[i].legend(loc='best', numpoints=1, fancybox=True,
                             ncol=2, bbox_to_anchor=(0.4, 1.15))
        if plotfile:
            # fig.tight_layout()
            fig.savefig(plotfile, bbox_inches="tight")


def plot_ms(fig, ms_on, ms_off, axisName, tbase):
    """Plot the ms on/off values as blue and cyan points, respectively."""
    axes = fig.get_axes()[0]
    ymin, ymax = axes.get_ylim()

    for ms, ctype, label in [(ms_on, 'blue', 'MS on'),
                             (ms_off, 'cyan', 'MS off')]:
        ms_time = np.array(
            [t for t, a in zip(ms["time"], ms_on["axis"])
             if a == axisName.upper()])
        ms_time = ms_time[ms_time > 0] - tbase
        y = np.zeros_like(ms_time) + ymin + 0.1*(ymax - ymin)

        axes.plot(ms_time, y, "+", ms=10, mew=3, color=ctype, label=label)


def do_scale(fiducials, fiducials_deg, axisName, pos, fpos, nfpos, fididx,
             verbose=False):
    """
    If the axis wraps, ensure that pairs of fiducials seen 360 degrees apart are
    always separated by the same amount -- i.e. estimate the axis' scale.
    """
    axis_scale = {}
    match = {}
    for n in ('pos1', 'pos2'):
        ndiff, diff = 0, 0.0
        for i in range(len(fiducials)):
            # skip missing ones
            if fiducials_deg[i] < -999 or not np.isfinite(fiducials_deg[i]):
                continue
            # tmp = fiducials[np.abs(
            # fiducials_deg - 360 - fiducials_deg[i]) < 1]  # warning
                # RuntimeWarning: invalid value encountered in less

            tmp = np.ndarray([0], dtype=int)
            for j in range(len(fiducials)):
                if fiducials_deg[j] < -999 or not np.isfinite(fiducials_deg[j]):
                    continue
                dd = abs(fiducials_deg[j] - 360 - fiducials_deg[i])
                if dd < 1:
                    fid = fiducials[j]
                    tmp = np.append(tmp, [fid])

            if len(tmp) == 0:
                continue            # we were only here once

            match[i] = tmp[0]
            # print "   --", fiducials[i], fiducials_deg[i],  match[i],
            #   fiducials_deg[match[i]] # , tmp

            tmp = pos[n][fididx == fiducials[i]]
            if len(tmp) == 0:
                continue            # we didn't cross this fiducial
            p1 = np.mean(tmp)

            tmp = pos[n][fididx == match[i]]
            # we didn't cross this fiducial
            if len(tmp) == 0:
                continue
            p2 = np.mean(tmp)

            diff += p2 - p1
            ndiff += 1

            # if verbose:
            #    print  fiducials[i], fiducials_deg[i], "--> ", match[i],
            #  fiducials_deg[match[i]]

        if ndiff == 0:
            if n == 1 and 'altitude' != axisName:
                print("You haven't moved a full 360degrees, so I cannot find"
                      " the scale", file=sys.stderr)

            for i in range(len(fiducials)):
                if nfpos[n][i] > 0:
                    v0 = fpos[n][i]
                    deg0 = fiducials_deg[i]
                    break

            for i in range(len(fiducials) - 1, -1, -1):
                if nfpos[n][i] > 0:
                    v1 = fpos[n][i]
                    deg1 = fiducials_deg[i]
                    break

            axis_scale[n] = 1.0/(v0 - v1)  # n.b. -ve; not a real scale
            if verbose:
                print("    %s scale APPROX %.6f  (encoder %s)" %
                      (axisName, 60.0*60.0*(deg0 - deg1)*axis_scale[n], n),
                      file=sys.stderr)
        else:
            diff /= ndiff
            axis_scale[n] = 60.0*60.0*360.0/diff
            if verbose:
                print("    %s scale = %.6f  (encoder %s)" %
                      (axisName, axis_scale[n], n[-1]))
            #
            # Force fiducials to have that scale
            #
            for i in range(len(fiducials)):
                if match.get(i):
                    mean = (fpos[n][i] + fpos[n][match[i]] - diff)/2.0  # error
                    fpos[n][i] = mean
                    fpos[n][match[i]] = mean + diff

    if verbose:
        print("    Encoder2/Encoder1 - 1 = %.6e" %
              (axis_scale['pos2']/axis_scale['pos1'] - 1))

    return axis_scale


def write_table_file(tableFile, fiducials, fpos, fposErr, nfpos,
                     axisName, dfile, axis_scale, ff, args):
    """Write the results to a fiducials table file for the MCP to use."""
    dfile = Path(dfile)
    ss = ""
    ss = ss+"-dir %s" % dfile.parent
    if args.mjd:
        ss = ss+" -mjd %i" % args.mjd
    ss = ss+" -%s" % axisName
    if args.t0:
        ss = ss+" -t0 %d" % args.t0
    if args.t1:
        ss = ss+" -t1 %d" % args.t1
    if args.setCanonical != None:
        ss = ss+" -setCanon %s" % args.setCanonical
    if args.canonical:
        ss = ss+" -canon"
    if args.resetFid != None:
        ss = ss+" -resetFid %s" % args.resetFid
    if args.reset:
        ss = ss+" -reset"
    if args.scale:
        ss = ss+" -scale"
    ss = ss+" -table %s" % args.tableFile

    if tableFile in ("stdout", "-",):
        fd = sys.stdout
    else:
        fd = open(tableFile, "w")

    # Get a CVS Name tag w/o writing it out literally.
    cvsname = "".join(["$", "Name", "$"])

    tt2 = Time.now().iso[:19] + 'Z'

    scale = {}
    for pos in ['pos1', 'pos2']:
        if axis_scale.get(pos) is None:
            scale[pos] = "None"
        else:
            scale[pos] = "%12.10f" % (axis_scale[pos])

    print("""#
# %s fiducials
#
# %s
#
# Creator:                 %s
# Time:                    %s
# Input file:              %s
# Scales:                  %s  %s
# Canonical fiducial:      %s
# Arguments:               %s
#
# Fiducial Encoder1 +-   error  npoint  Encoder2 +-   error  npoint
""" % (axisName, cvsname, pwd.getpwuid(os.geteuid())[0],
       tt2, dFileTail, scale['pos1'], scale['pos2'], ff, ss), end=' ', file=fd)

    for i in range(1, len(fpos['pos1'])):
        print("%-4d " % fiducials[i], end=' ', file=fd)

        for n in ('pos1', 'pos2'):
            print("   %10.0f +- %7.1f %3d" %
                  (fpos[n][i], fposErr[n][i], nfpos[n][i]), end=' ', file=fd)
        print("", file=fd)

    if fd != sys.stdout:
        fd.close()


def main(argv=None):
    """Read and analyze an mcpFiducials file,  e.g.
    plotMcpFiducials -mjd 51799 --alt -reset -canon -updown -table stdout
    """
    parser = argparse.ArgumentParser("plotMcpFiducials")
    parser.add_argument("fileName", type=str, nargs="?",
                        help="The mcpFiducials file to read")
    parser.add_argument("--azimuth", action="store_true",
                        default=False, help="Read azimuth data")
    parser.add_argument("--altitude", action="store_true",
                        default=False, help="Read altitude data")
    parser.add_argument("--rotator", action="store_true",
                        default=False, help="Read rotator data")

    parser.add_argument("--mjd", default=getMJD(), type=int,
                        help="Read mcpFiducials file for this MJD")
    parser.add_argument("-d", "--dir", default=sdss_paths.mcp_logs,
                        dest="mcp_log_dir",
                        help="Directory to search for input file")

    parser.add_argument("--index0", default=0, type=int,
                        help="Ignore all data in the mcpFiducials file with"
                        " index < index0")
    parser.add_argument("--index1", default=0, type=int,
                        help="Ignore all data in the mcpFiducials file with"
                        " index > index0")

    parser.add_argument("--t0", type=int, default=0,
                        help="Ignore all data in the mcpFiducials file older"
                        " than this timestamp; If less than starting time,"
                        " interpret as starting time + |t0|. See also --index0")
    parser.add_argument("--t1", type=int, default=0,
                        help="""Ignore all data in the mcpFiducials file younger
                        than this timestamp; If less than starting time,
                        interpret as starting time + |t1|. See also
                        --index1""")

    parser.add_argument("--canonical", action="store_true", default=False,
                        help="Set the `canonical' fiducial to its canonical"
                        " value")
    parser.add_argument("--setCanonical", type=str, default=None,
                        help="Specify the canonical fiducial and optionally"
                        " position, e.g. --setCanonical 78 --setCanonical"
                        " 78:168185")
    parser.add_argument("--error", action="store_true", default=False,
                        help="Plot error between true and observed fiducial"
                        " positions")
    parser.add_argument("--errormax", type=float, default=0.0,
                        help="Ignore errors between true and observed fiducial"
                        " positions larger than <max>")
    parser.add_argument("--reset", action="store_true", default=False,
                        help="Reset fiducials every time we recross the first"
                        " one in the file?")
    parser.add_argument("-resetFid",  type=int,
                        help="reset fiducial index if default oe is not good")
    parser.add_argument("--scale", action="store_true", default=False,
                        help="Calculate the scale and ensure that it's the same"
                        " for all fiducial pairs?")
    parser.add_argument("--absolute", action="store_true", default=False,
                        help="Plot the values of the encoder, not the scatter"
                        " about the mean")
    parser.add_argument("--updown", action="store_true", default=False,
                        help="Analyze the `up' and `down' crossings separately")

    parser.add_argument("--ms", action="store_true", default=False,
                        help="Show MS actions on the time axis")
    parser.add_argument("--noplot", action="store_true",
                        help="Don't generate any plots.")
    parser.add_argument("--plotFile", action="store_true",
                        help="Write plot to a file (name automatically"
                        " determined).")
    parser.add_argument("--tableFile", type=str, default='',
                        help='Write fiducials table to this file (may be'
                        ' "stdout", or "-")')
    parser.add_argument("-v", "--verbose", action="store_true", default=False,
                        help="Be chatty?")

    parser.add_argument("--read_old_mcpFiducials", action="store_true",
                        default=False,
                        help="Read old mcpFiducial files (without encoder2"
                        " info)?")

    args = parser.parse_args(argv)

    if args.verbose:
        print("---------------------")
        print(f"Program: {Path(__file__).absolute()}")


# input verification
    #
    # Check that flags make sense
    #
    if args.error:
        if args.canonical:
            print("Please don't specify -canonical with -error", file=sys.stderr)
            sys.exit(1)
        if args.reset:
            print("Please don't specify -reset with -error", file=sys.stderr)
            sys.exit(1)
        if args.updown:
            print("Please don't specify -updown with -error", file=sys.stderr)
            sys.exit(1)
        if args.tableFile:
            print("Please don't specify e.g. -table with -error",
                  file=sys.stderr)
            sys.exit(1)


# -- em -- Scale + absolute
    if args.scale:
        if not args.absolute:
            if args.verbose:
                print("-scale only makes sense with -absolute; I'll set it for"
                      " you")
            args.absolute = True

#  --em -- Pyplot? ------
    if pyplot is None:
        print("I am unable to plot as I failed to import matplotlib",
              file=sys.stderr)

# --em -  AXIS -------
    # Which axis are we interested in?
    naxis = args.azimuth + args.altitude + args.rotator

    if not naxis:
        print("Please specify an axis", file=sys.stderr)
        sys.exit(1)
    elif naxis > 1:
        print("Please specify only one axis", file=sys.stderr)
        sys.exit(1)

    if args.altitude:
        axisName = "altitude"
        struct = "ALT_FIDUCIAL"
        names = ["time", "fididx", "pos1",
                 "pos2", "deg", "alt_pos", "velocity"]
    if args.azimuth:
        axisName = "azimuth"
        struct = "AZ_FIDUCIAL"
        names = ["time", "fididx", "pos1", "pos2", "deg", "velocity"]
    if args.rotator:
        axisName = "rotator"
        struct = "ROT_FIDUCIAL"
        names = ["time", "fididx", "pos1", "pos2", "deg", "latch", "velocity"]

# --- old_mcpFiducials ?
    if args.read_old_mcpFiducials:
        names.append("true")
    else:
        names.append("true1")
        names.append("true2")

    if args.verbose:
        print("Axis name=%s, struct=%s, names=%s" % (axisName, struct, names))

# --  get the name of data file
    dfile = get_input_file(args.fileName, args.mjd,
                           args.mcp_log_dir, verbose=args.verbose)
    if args.verbose:
        print("Input file =", dfile)

# --- em Read input file to VECS structure, using yanny function
    if".gz" in dfile.suffix:
        tmp_file = Path.home() / dfile.with_suffix("").name
        print(tmp_file)
        with gzip.open(dfile, "rb") as in_fil:
            with tmp_file.open("w") as out_fil:
                out_fil.write(in_fil.read().decode("utf-8"))
        pars = yanny(tmp_file.open('r'))
        tmp_file.unlink()
    else:
        pars = yanny(dfile)
    vecs = pars[struct]

    if args.verbose:
        print("Input vecs, len(vecs)=%s" % len(vecs))


# -----   ms on/off  ?
    ms = {}
    if args.ms:
        ms['on'] = pars["MS_ON"]
        ms['off'] = pars["MS_OFF"]

# em ------------------------------------------
    # Discard early data if so directed
    t0, t1 = args.t0, args.t1
    index0, index1 = args.index0, args.index1
    tmod = 10000
    tbase = tmod*int(vecs['time'][0]/tmod)

    def rescale_time(t0):
        if t0 != 0 and t0 < vecs['time'][0]:
            t0 = tbase + abs(t0)

    rescale_time(t0)
    rescale_time(t1)

    if t0 != 0 or t1 != 0 or index0 != 0 or index1 != 0:
        if t0 != 0 or t1 != 0:
            if index0 != 0 or index1 != 0:
                print(
                    "Ignoring index[01] in favour of time[01]", file=sys.stderr)

            if t0 == 0:
                t0 = vecs['time'][0]
            if t1 == 0:
                t1 = vecs['time'][-1]

            tmp = (vecs['time'] >= t0) & (vecs['time'] <= t1)
        else:
            index = list(range(0, len(vecs['time'])))
            if index0 < 0:
                index0 = 0
            if index1 == 0 or index1 >= len(vecs['time']):
                index1 = len(vecs['time']) - 1

            tmp = index >= index0 and index <= index1

        vecs = vecs[tmp]

    # re-create the index array into the newly-truncated data.
    index = np.arange(len(vecs['time']))

    # Are there any datavalues left?
    if len(vecs['time']) == 0:
        print("No points to plot after applying the time arguments, exit",
              file=sys.stderr)
        sys.exit(1)

    # Reduce time to some readable value
    vecs['time'] -= tbase

    # re-create the index array into the newly-truncated data.
    index = np.arange(len(vecs['time']))
    if args.verbose:
        print("Applied time frame:  len(vecs)={}".format(len(vecs)))

    # Discard the non-mark rotator fiducials
    if args.rotator:
        tmp = vecs['fididx'] > 0
        vecs = vecs[tmp]
        index = np.arange(len(vecs))
        # re-create the index array into the newly-truncated data.
        index = np.arange(len(vecs['time']))
        if args.verbose:
            print("Discard the non-mark rotator fiducials:  len(vecs)={}"
                  "".format(len(vecs)))

# ---------------------------
    # If we want to analyse the `up' and `down' fiducials separately,
    # add ud_delta to fididx for all crossings with -ve velocity
    if args.updown:
        ud_delta = max(vecs['fididx']) + 5
        vecs['fididx'][vecs['velocity'] < 0] += ud_delta
    else:
        ud_delta = 0

# ------ Processing
    # Process fiducial data
    # Start by finding the names of all fiducials crossed

#  E<: JP has "True", and we had fiducials starting from 0. Changed to False.
#  Code from proc:
    # fiducials = make_fiducials_vector(vecs['fididx'], False)
    #  #  EM: np.array(sorted(set(fididx)),dtype=int)
    # JP : np.arange(0, max(fididx)+1)
    fiducials = make_fiducials_vector(vecs['fididx'], True)

    # Find the approximate position of each fiducial (in degrees)
    fiducials_deg = np.empty_like(fiducials) + np.nan
    for i in range(len(fiducials)):
        tmp = vecs['deg'][vecs['fididx'] == fiducials[i]]
        if tmp != []:
            # fiducials_deg[i] = tmp[0]   # RL verbose output match with this definition
            fiducials_deg[i] = np.mean(tmp)

    if args.verbose:
        print("Find the approximate position of each fiducial (in degrees)")
        print("Fididx  Degrees    N   Pos1")
        for i in range(1, len(fiducials)):
            # if np.isfinite(fiducials_deg[i]):  # why John used this?
            tmp2 = vecs['pos1'][vecs['fididx'] == fiducials[i]]
            if len(tmp2) == 0:
                ss = np.nan
            else:
                ss = np.mean(tmp2)
            print(" %3d\t%8.3f  %2d  %10.0f " %
                  (fiducials[i], fiducials_deg[i], len(tmp2), ss))

# L409
    # Unless we just want to see the fiducial errors (-error), estimate
    # the mean of each fiducial value
    fpos = {}
    fpos['pos1'] = np.zeros_like(fiducials)
    fpos['pos2'] = np.zeros_like(fiducials)

    if args.error:
        # Read the fiducial file if provided
        if args.fiducialFile:
            tempvec = read_fiducials(args.fiducialFile, axisName, False)[1]
            for pos in ('pos1', 'pos2'):
                for i, fid in enumerate(fiducials):
                    tmp = tempvec[pos][tempvec['findex'] == fid]
                    fpos[pos][i] = tmp[0] if len(tmp) > 0 else 0
        else:
            # Find the true value for each fiducial we've crossed
            if args.read_old_mcpFiducials:
                true1 = True
            for i in range(len(fiducials)):
                fpos['pos1'][i] = np.mean(
                    true1[vecs['fididx'] == fiducials[i]])
            fpos['pos2'] = fpos['pos1']

    else:
        # --em ------ RESET------
        # Reset the encoders every time that they pass some `fiducial' fiducial?
        if args.reset:
            reset_fiducials(vecs['fididx'], ud_delta, vecs['velocity'],
                            vecs['pos1'], args.verbose, axisName, args.resetFid)
            reset_fiducials(vecs['fididx'], ud_delta, vecs['velocity'],
                            vecs['pos2'], False, axisName, args.resetFid)
            # don't double-print

# -- em ------
        # Calculate mean values of pos[12]
        if args.updown:
            pos0 = {}
            for pos in ('pos1', 'pos2'):
                pos0[pos] = vecs[pos]       # save pos[pos]

                for i, fid in enumerate(fiducials):
                    fpos[pos][i] = np.mean(
                        vecs[pos][vecs['fididx'] == fid])

                down = np.where(vecs['fididx'] < ud_delta)
                tmp2 = vecs['fididx']
                tmp2[down] += ud_delta
                tmp[np.logical_not(down)] -= ud_delta

                tmp = fpos[pos][vecs['fididx']] + fpos[pos][tmp2]
                tmp = tmp/((fpos[pos][vecs['fididx']] != 0) +
                           (fpos[pos][tmp2] != 0))

                vecs[pos] -= fpos[pos][vecs['fididx']] - tmp
            # Undo that +$ud_delta
            vecs['fididx'][vecs['fididx'] > ud_delta] -= ud_delta

            fiducials = make_fiducials_vector(vecs['fididx'])

            for pos in ('pos1', 'pos2'):
                fpos[pos] = np.zeros_like(fiducials)
        # --------

        # (Re)calculate fpos[12] --- `Re' if args.updown was true

        def recalc_pos_err_nfpos(vecs, fiducials):
            fposErr = {}
            nfpos = {}
            fpos = {}
            for pos in ('pos1', 'pos2'):
                nfpos[pos] = np.empty_like(fiducials)
                fposErr[pos] = np.empty_like(fiducials, dtype=float)
                fpos[pos] = np.zeros_like(fiducials, dtype=float)

            for i in range(len(fiducials)):
                for pos in ('pos1', 'pos2'):
                    tmp = vecs[pos][vecs['fididx'] == fiducials[i]]
                    if len(tmp) == 0:
                        fpos[pos][i] = 0
                        nfpos[pos][i] = 0
                        fposErr[pos][i] = -9999  # disable fiducial
                    else:
                        fpos[pos][i] = np.mean(tmp)
                        nfpos[pos][i] = len(tmp)
                        # changed statistics np.std(tmp) -->
                        #   np.std(tmp)/np.sqrt(nfpos[pos][i]-1)
                        fposErr[pos][i] = 1e10 if len(tmp) == 1 else np.std(
                            tmp)/np.sqrt(nfpos[pos][i]-1)
            return fpos, nfpos, fposErr

        fpos, nfpos, fposErr = recalc_pos_err_nfpos(vecs, fiducials)

        #fposErr = {}; nfpos = {}
        # for pos in ('pos1','pos2'):
        #    nfpos[pos] = np.empty_like(fiducials)
        #    fposErr[pos] = np.empty_like(fiducials)

        # for i in range(len(fiducials)):
        #    for pos in ('pos1','pos2'):
        #        tmp =  vecs[pos][vecs['fididx'] == fiducials[i]]
        #        print i, pos, "tmp=", tmp
        #        if len(tmp) == 0:
        #            fposErr[pos][i] = -9999 # disable fiducial
        #            nfpos[pos][i] = 0
        #        else:
        #            fpos[pos][i] = np.mean(tmp)
        #            nfpos[pos][i] = len(tmp)
        #            fposErr[pos][i] = 1e10 if len(tmp) == 1 else np.std(tmp)

        # restore values
        if args.updown:
            for pos in ('pos1', 'pos2'):
                vecs['pos'][pos] = pos0[pos]

# ------ absolute ----------
    # absolute vecs is using for scale, and relative is using for errors
    # default args.absolute is False,
    # if args.scale,  set args.absolute=True
    if not args.absolute:
        if args.verbose:
            print(
                "args.absolute={}, change vecs[pos] to relative values".format(
                    args.absolute))
        for pos in ('pos1', 'pos2'):
            vecs[pos] = vecs[pos] - fpos[pos][vecs['fididx']]
            # next is working for not extended fiducials vector-em
            # for i,ind in enumerate(vecs['fididx']):
            #    i0=np.where(fiducials == ind)
            #    vecs[pos][i] = vecs[pos][i] - fpos[pos][i0[0][0]]
    # else:
    #    if args.verbose:
    #        print  "args.absolute=%s, keep vecs[pos] absolue values" % args.absolute

# -------- SCALE--------
    axis_scale = {}
    if args.scale:
        if args.verbose:
            print("Calculating scale: ")
        axis_scale = do_scale(fiducials, fiducials_deg, axisName,
                              vecs, fpos, nfpos, vecs['fididx'],
                              verbose=args.verbose)

# ------- canonical ---------------------
    # Fix fiducial positions to match canonical value for some fiducial?
    ff = None
    if args.canonical:
        ff, ffval, ss = set_canonical_fiducials(
            axisName, vecs, args.setCanonical)
        if args.verbose:
            print("Applying canonical fiducial value:  %s:%s  %s  " %
                  (ff, ffval, ss))
        for pos in ('pos1', 'pos2'):
            tmp = fpos[pos][fiducials == ff]
            # tmp = tmp[tmp != 0]     # valid crossings aren't == 0?   - reset
            #  fid has zero encod value
            if len(tmp) == 0:
                print("   ------??? Error: You have not crossed/read canonical"
                      " fiducial %d" %
                      (ff), file=sys.stderr)
                sys.exit(1)
            # fpos[pos][nfpos[pos] > 0] += np.int64(ffval - np.mean(tmp))  # JP
            fpos[pos][nfpos[pos] > 0] += (ffval - tmp[0])  # em

# -------   round up --- em --
    for pos in ('pos1', 'pos2'):
        fpos[pos] = np.rint(fpos[pos])
        fposErr[pos] = np.around(fposErr[pos], decimals=1)

# -------- write table file --------
    # Print table?
    if args.tableFile:
        if args.verbose:
            print("Output to %s" % args.tableFile)
        write_table_file(args.tableFile, fiducials, fpos, fposErr, nfpos,
                         axisName, dfile, axis_scale, ff, args)  # it was argv

# -------- args.error -------
    if args.error:
        # we have a measurement of this fiducial
        haveValue = np.where(fpos['pos1'][vecs['fididx']] != 0)
        if args.errormax > 0:
            haveValue = haveValue & abs(vecs['pos1']) < args.errormax
        vecs = vecs[haveValue]

# -------- plot/noplot ------
    if not args.noplot:
        # plot_data(axisName, vecs, index, args.plotFile, ffile, ms, args.reset,
        #  args.updown, args.error, t0, tbase)
        plot_data(axisName, vecs, index, args.plotFile, dfile, ms,
                  args.reset, args.updown, args.error, t0, tbase)
        input("Press enter to exit... ")

    return fiducials, fpos, fposErr, nfpos


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from pathlib import Path
from astropy.time import Time
import multiprocessing

from bin import sjd, influx_fetch

try:
    import tpmdata
except ImportError:
    tpmdata = None

__version__ = "3.0.0"


def get_tpm_packet(out_dict):
    tpmdata.tinit()
    data = tpmdata.packet(1, 1)
    for key, val in data.items():
        out_dict[key] = val
    return 0

def query():
    
    t_start = Time(sjd.sjd() - 0.3, format="mjd")
    t_end = Time.now()
    enclosure_path = Path(__file__).parent.parent / "flux/enclosure.flux"
    state = influx_fetch.query(enclosure_path.open('r').read(), t_start, t_end)[0]
    enclosure_hist = ""
    last_state = 0
    for row in state.records:
        if row.get_value() > last_state:
            last_state = row.get_value()
            t = Time(row.get_time())
            enclosure_hist += f"Opened at {t.isot[11:19]}\n"
        elif row.get_value() < last_state:
            last_state = row.get_value()
            t = Time(row.get_time())
            enclosure_hist += f"Closed at {t.isot[11:19]}\n"
    if enclosure_hist == "":
        enclosure_hist = "Closed all night\n"
            
        
    if tpmdata is None:
        raise ConnectionError("Cannot query the tpm without tpmdata installed")
    
    data = multiprocessing.Manager().dict()
    tpm_thread = multiprocessing.Process(target=get_tpm_packet, args=(data,))
    tpm_thread.start()
    tpm_thread.join(2)
    if tpm_thread.is_alive():
        tpm_thread.kill()
        raise ConnectionError("Could not reach TPM")

    t = Time(data["ctime"], format="unix")
    output = enclosure_hist + '\n'
    output += f"Status at:  {t.isot[12:19]}Z\n"
    output += (f"Telescope Position:  "
               f"{data['az_actual_pos']*data['az_spt']/3600:>5.1f},"
               f" {data['alt_actual_pos']*data['alt_spt']/3600:>5.1f},"
               f" {data['rot_actual_pos']*data['rot_spt']/3600:>5.1f} mount\n")
    # epics_data = epics_fetch.get_data(["25m:mcp:instrumentNum"],
                                    #   start_time=Time.now().to_datetime(),
                                    #   end_time=Time.now().to_datetime())
    cart = "FPS" if data["inst_id_0"] == 0 else f"{data['inst_id_0']:.0f}"
    output += f"Instrument mounted:  {cart}\n"
    output += (f"Counterweights at:  {data['plc_cw_0']:.1f},"
               f" {data['plc_cw_1']:.1f},"
               f" {data['plc_cw_2']:.1f},"
               f" {data['plc_cw_3']:.1f}\n")
    # TODO Check if this is supposed to do something
    if data["dewar_sp1_psi"] > 10:
        output += f"LN2 autofill systems:  Connected and turned on\n"
    else:
        output += f"LN2 autofill systems: Disconnected\n"
    output += (f"180L LN2 dewar scale:  SP1 {data['dewar_sp1_lb']:6.1f} lbs,"
               f" {data['dewar_sp1_psi']:6.1f} psi")

    return output


def main():
    print(query())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import unittest
from pathlib import Path
import subprocess as sub
from astropy.time import Time
from bin import influx_fetch
    
class TestInfluxFetch(unittest.TestCase):
    def test_known_date(self):
        dust = Path(__file__).parent.parent / "flux/dust.flux"
        with dust.open('r') as fil:
            query = fil.read()
        influx_fetch.query(query, Time.now(), Time.now() - 1, interval="1h")

    def test_archive(self):
        """Checks to see if the directory for new data is available to this
        computer"""
        client = influx_fetch.get_client()
        print(client.ready())

    def test_help(self):
        """"Prints the help if -h is provided"""
        sub.call('{} -h'.format(epics_fetch.__file__), shell=True)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
import unittest
from pathlib import Path
from bin import ap_test
import numpy as np
from sdssobstools import sdss_paths


class TestAPTest(unittest.TestCase):
    def setUp(self) -> None:
        self.project = Path(__file__).absolute().parent.parent

        class Dummy(object):
            pass

        self.args = Dummy()
        self.args.sjds = [59593]
        self.args.exps = [[40310015, 40310016, 40310017, 40310018, 40310019]]
        self.args.plot = False
        self.args.verbose = True

    def test_known_date(self):

        self.atest = ap_test.ApogeeFlat(self.args)
        self.atest.run_inputs()

    def test_directory(self):
        """Checks to see if the directory for new data is available to this
        computer"""
        self.assertTrue(sdss_paths.ap_utr.exists())

    def test_plotting(self):
        """Tests the plotting routine"""
        self.args.plot = True
        self.atest = ap_test.ApogeeFlat(self.args)
        self.atest.run_inputs()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""A script to process gfa images with the SDSS-V FPS"""
import tqdm
import fitsio
import sep
import argparse
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from pathlib import Path
from matplotlib.patches import Ellipse

from bin import sjd
from sdssobstools import sdss_paths

sns.set(style="darkgrid")

__author__ = "Dylan Gatlin"


def build_filt(obj_arr: np.ndarray):
    ecc = np.sqrt(obj_arr['a']**2 - obj_arr['b']**2) / obj_arr['a']
    # print(np.percentile(ecc, 75))
    # print(obj_arr["cpeak"], obj_arr["peak"].max())
    # print(ecc, np.mean(ecc), np.percentile(ecc, 95))
    filt = (
        # (obj_arr["flux"] > np.percentile(obj_arr["flux"], 80))
        ((obj_arr["a"] * 0.216) > 0.5)
        & (obj_arr["cpeak"] < 60000)
        # & (obj_arr["npix"] > 5)
        & (ecc < 0.7)
        )
    # print(obj_arr["a"] * 0.216, obj_arr["b"] * 0.216,
    # np.mean([obj_arr["a"], obj_arr["b"]]))
    return filt

def show_img(obj_arr: np.ndarray, obj_filt, data_sub: np.ndarray, plot_file=""):
    fig, ax = plt.subplots(1, 1, figsize=(8, 8))
    m, s = np.mean(data_sub), np.std(data_sub)
    im = ax.imshow(data_sub, interpolation='nearest', cmap='gray',
                   vmin=m-s, vmax=m+s, origin='lower')

    # plot an ellipse for each object
    for i, obj in enumerate(obj_arr[obj_filt]):
        e = Ellipse(xy=(obj['x'], obj['y']),
                    width=6*obj['a'],
                    height=6*obj['b'],
                    angle=obj['theta'] * 180. / np.pi)
        e.set_facecolor('none')
        e.set_edgecolor('red')
        ax.add_artist(e)
    plt.show()
    if plot_file:
        fig.savefig(plot_file)



class GFASet:
    def __init__(self, verbose=False, exp_num_plot=False):
        self.paths = []
        self.filter = []
        self.isots = []
        self.fwhms = []
        self.focuses = []
        self.n_objs = []
        self.im_nums = []
        self.verbose = verbose
        self.exp_num_plot = exp_num_plot

    def add_index(self, paths_iter, im_num):
        paths = []
        filter = []
        focus = []
        isots = []
        fwhms = []
        n_objs = []
        for p in paths_iter:
            paths.append(p)
            if p.exists():
                filter.append(True)
                fwhm, n, objs = self.get_fwhm(p, verbose=self.verbose)
                hdr = fitsio.read_header(p, 1)
                focus.append(hdr["FOCUS"])
                fwhms.append(fwhm)
                n_objs.append(n)
            else:
                filter.append(False)
                focus.append(np.nan)
                fwhms.append(np.nan)
                n_objs.append(0)
        self.im_nums.append(im_num)
        self.paths.append(paths)
        self.focuses.append(focus)
        self.filter.append(filter)
        self.fwhms.append(fwhms)
        self.n_objs.append(n_objs)

    @staticmethod
    def get_fwhm(path: Path, verbose: bool=False):
        data = fitsio.read(path, 1)
        bkg = sep.Background(data.astype(float))
        objs = sep.extract(data - bkg, 1.5, bkg.globalrms)
        filt = build_filt(objs)
        if verbose:
            print(f"Using {filt.sum()}/{filt.shape[0]} points")
        # fwhm = 0.216 * np.mean(2.0*np.sqrt(
            # 2.0*np.log(2)*(objs['a'][filt]**2 + objs['b'][filt]**2))
        # )
        fwhm = 0.216 * np.mean(
             np.sqrt(2 *  np.log(2) * objs['a'][filt]**2 + objs['b'][filt]**2)
            )
        return fwhm, filt.sum(), objs

    def sort(self):
        self.im_nums = np.array(self.im_nums)
        self.paths = np.array(self.paths)
        self.focuses = np.array(self.focuses)
        self.filter = np.array(self.filter)
        self.fwhms = np.array(self.fwhms)
        self.n_objs = np.array(self.n_objs)
        sorter = self.im_nums.argsort()
        self.im_nums = self.im_nums[sorter]
        self.paths = self.paths[sorter]
        self.focuses = self.focuses[sorter]
        self.filter = self.filter[sorter]
        self.fwhms = self.fwhms[sorter]
        self.n_objs = self.n_objs[sorter]

    def print(self):
        if self.verbose:
            print("Measurements in arcseconds with a place scale of 0.216")
        print(f"{'Img #':<6} {'GFA1':<6} {'GFA2':<6} {'GFA3':<6} {'GFA4':<6}"
            f" {'GFA5':<6} {'GFA6':<6} {'N Objs':<6}")
        print('=' * 80)
        for i, im, in enumerate(self.im_nums):
            print(f"{im:>6} "
                + " ".join([f"{f:>6.2f}"
                    if not np.isnan(f) else f"{'-':>6}"
                    for (j, f) in enumerate(self.fwhms[i])]
                ) + f" {np.mean(self.n_objs[i]):>6.0f}"
            )
    
    @staticmethod
    def quadratic(x, a, b, c): 
        return a * x**2 + x * b + c

    def plot(self, plot_file):
        if not self.exp_num_plot:
            flat_foc = self.focuses.flatten()
            flat_fwhms = self.fwhms.flatten()
            nan_filt = ~np.isnan(flat_fwhms)
            a, b, c = np.polyfit(flat_foc[nan_filt],
                                 flat_fwhms[nan_filt], deg=2)
            focs = np.linspace(flat_foc[nan_filt].min(),
                               flat_foc[nan_filt].max(), 100)
            fit = -b / 2 /a
            fwhm = self.quadratic(fit, a, b, c)
            mum = "\u03BCm"
            if a > 0:
                fit_found = False
                print("Optimal focus not found")
            else:
                fit_found = True
                print(f'Optimal Focus is at {fit:.0f}{mum} with Width {fwhm:.1f}"')
        fig, ax = plt.subplots(1, 1, figsize=(6, 4))
        if (not self.exp_num_plot) and fit_found:
            ax.plot(focs, self.quadratic(focs, a, b, c),
                    label=f"Quad Fit, best={-b / 2 / a:.0f}", alpha=0.8)
            ax.axvline(-b / 2 / a, c="r", linestyle="--", alpha=0.6)
        for i in range(6):
            if self.exp_num_plot:
                ax.plot(self.im_nums, self.fwhms[:, i], linewidth=1, alpha=0.8,
                label=f"GFA {i+1}")
                ax.scatter(self.im_nums, self.fwhms[:, i], s=6, alpha=0.8)
            else:
                ax.scatter(self.focuses[:, i], self.fwhms[:, i], s=6, alpha=0.8,
                label=f"GFA {i+1}")
        ax.legend()
        if self.exp_num_plot:
            ax.set_xlabel("Exposure Number")
        else:
            ax.set_xlabel("Focus ($\mu m$)")
        ax.set_ylabel("FWHM (arcseconds)")
        plt.show()
        if plot_file:
            fig.savefig(plot_file, dpi=100)


def parse_args():
    parser = argparse.ArgumentParser(description="Prints a table of FWHMs from"
        " GFAs, useful for focus sweeps")
    parser.add_argument("-b", "--bias", help="The index of a bias frame")
    parser.add_argument("-m", "--mjd", default=sjd.sjd())
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Verbose debugging output")
    parser.add_argument("-f", "--file", nargs="+",
                        help="Specific files to process")
    parser.add_argument("-n", "--gfas", nargs="+", type=int, default=[2, 5])
    parser.add_argument("-w", "--window",
                        help="A dash separated beginning and end of the window."
                             " Always include a --master-field with your window."
                             " Ex: -w 15-50 -s 35")
    parser.add_argument("-e", "--exp-num", action="store_true",
                        help="Don't fit the images to a curve,"
                        " produce a time series across exposure numbers")
    # parser.add_argument("-s", "--master-field", type=int,
    #                     help="The master field with all stars you want. Only"
    #                     " used in conjunction with --window. You need both"
    #                     " arguments.")
    parser.add_argument("-p", "--plot", action="store_true",
                        help="Show plot to analyze fit quality")
    parser.add_argument("-k", "--plot-file", help="Filename to save plot to."
                        " only works with --plot and --window together. If not"
                        " included, the plot will show up in a new window (or"
                        " maybe not depending on X11).")
    parser.add_argument("-d", "--plot-image", action="store_true",
        help="Plot an the sky image with ellipses traced, only works with -f"
    )
    args = parser.parse_args()

    return args


def main(args=None):
    if args is None:
        args = parse_args()
    gfas = GFASet(verbose=args.verbose, exp_num_plot=args.exp_num)
    if args.bias:
        raise NotImplementedError("This script doesn't support bias reduction")
    if args.file:
        print(f"{'File Name':<20} {'FWHM':<6} {'N Objects':<11} {'Mean Ecc':<6}")
        print('=' * 80)
        for file in args.file:
            if '*' not in file:
                file = Path(file)
                if not file.exists():
                    raise FileNotFoundError(f"Couldn't find {file.absolute()}")
                fwhm, n_objs, objs = gfas.get_fwhm(file)
                filt = build_filt(objs)
                ecc = (np.sqrt(objs['a'][filt]**2 - objs['b'][filt]**2)
                    / objs['a'][filt])
                print(f"{file.name:<20} {fwhm:>6.2f} {filt.sum():>5.0f}"
                    f"/{len(objs):<5.0f}  {np.mean(ecc):>6.2f}")
                if args.plot_image:
                    data = fitsio.read(file, 1)
                    bkg = sep.Background(data.astype(float))
                    show_img(objs, filt, data - bkg)
            else:
                file = Path(file)
                for p in file.parent.glob(file.name):
                    pass
    elif not args.window:
        path_stem = sdss_paths.gcam / f"{args.mjd}"
        low = 1000
        high = 0
        for file in path_stem.glob("gimg-gfa5n-*.fits*"):
            ind = int(file.name.split("-")[-1].rstrip(
                ".fits").rstrip(".fits.gz"))
            if ind < low:
                low = ind
            if ind > high:
                high = ind
        if high > low:
            args.window = f"{low:.0f}-{high:.0f}"

    if args.window:
        low, high = args.window.split("-")
        try:
            low = int(low)
            high = int(high)
        except ValueError:
            raise ValueError(f"Window range must be two ints with a - between"
                             f" them {args.window}")
        if args.verbose:
            print(f"Reading Images from {low} to {high} in MJD {args.mjd}")
        for im_num in tqdm.tqdm(range(low, high + 1)):
            im_ps = []
            for n in range(1, 7):
                p = (sdss_paths.gcam / f"{args.mjd}/proc-gimg-gfa{n:.0f}n-"
                     f"{im_num:0>4.0f}.fits")
                if not p.exists():
                    p = (sdss_paths.gcam / f"{args.mjd}/proc-gimg-gfa{n:.0f}n-"
                         f"{im_num:0>4.0f}.fits.gz")
                if not p.exists():
                    p = (sdss_paths.gcam / f"{args.mjd}/gimg-gfa{n:.0f}n-"
                         f"{im_num:0>4.0f}.fits")
                if not p.exists():
                    p = (sdss_paths.gcam / f"{args.mjd}/gimg-gfa{n:.0f}n-"
                         f"{im_num:0>4.0f}.fits.gz")
                im_ps.append(p)

            gfas.add_index(im_ps, im_num)
        gfas.sort()
        gfas.print()
        if args.plot:
            gfas.plot(args.plot_file)
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python

'''apogeeThar.py  check apogee Thar arcs (EM)

Program to check the stability of APOGEE instrument:
 - take ArcLamp THAR quick-red data 
 - takes 1-3 spectra  which is optional
 - do fitting of gaussian function to one spectral line on each chip.
 - compare results with reference, average for mjd=56531, Aug 26,2013, 
             ff=09690003, 09690005, 09690014, 09690016.
  
EM 09/01/2013
Usage:  ./apogeeThar.py <-m1 mjd1> <-m2 mjd2> 

examples:
 ./apogeeThar.py -m1 56531    # reference
 ./apogeeThar.py -m1 56560 -m2  56567
 ./apogeeThar.py -m1 56445  -m2 56537    # 06/01/2013 (56445) -- 09/01/2013 (56537)

History: 
09/11/2013: EM added to apogeeql svn bin repository
09/16/2013: EM reference is the first calibration set after 2013 summer
   mjd=56531, Aug 26,2013, ff=09690003, 09690005, 09690014, 09690016     
09/30/2013: EM:   use quick-red instead of instead of raw data; 
   reorganized output for differences only, 3-lines and 3 rows output, 
   designed for night log. 
09/02/2014  switchehd to sdss-4,  needed more libraries, commented 
    matplotlib for a while until installed.
06/17/2020: DG: Changed sort to sorted (sort doesn't exist)
       
 '''
import argparse
import numpy, scipy
from astropy.io import fits
#from pylab import *
import scipy.optimize
import sys, os.path, glob
import time

#  Constants
#--p0a = scipy.c_[53864, 939.646, 1.2745]
p0b = scipy.c_[46184.2, 924.366, 1.071]
p0c = scipy.c_[31715, 1776.62, 0.803]
p0a = scipy.c_[53864, 939.646, 1.2745, 40.]

zone=20
path="/data/apogee/quickred/"

#---------  
def sdth(dth): 
    if dth==12.994:  return "A"
    elif dth==13.499:  return "B"
    else: return "?"    
#-------
def getFileName(path, mjd, chip, fExp="*"):
   return "%s%s/ap1D-%s-%s.fits.fz" % (path, mjd, chip,fExp)       
#------------------------
def curSjd():
# current mjd
  TAI_UTC =34; sjd1=(time.time() + TAI_UTC) / 86400.0 + 40587.3;  
  sjd= int (sjd1)
  return sjd
#----------------------- 

def OneFileFitting(data1, fiber, pRef):
  p0=pRef.copy()
# select zone around line  center
  x1=p0[0][1]-zone;   x2=p0[0][1]+zone
  x=numpy.arange(data1.shape[1])[x1:x2]  #  x-axis array in pix
  spe=data1[fiber,x1:x2]  #read spectrum in the line range 

  ll=numpy.where(spe == max(spe) )
  p0[0][1]= ll[0][0]+x1   

  fitfunc = lambda p0, x: p0[0]*scipy.exp(-(x-p0[1])**2/(2.0*p0[2]**2))
  errfunc = lambda p, x, y: fitfunc(p,x)-y
  p1, success= scipy.optimize.leastsq(errfunc, p0.copy()[0],args=(x,spe))
  
  ref= fitfunc(pRef[0], x)
  fit= fitfunc(p1, x)
  return success, p1, x, spe, ref, fit


def OneFileFitting2(data1, fiber, pRef, xmin_fit, xmax_fit):
  p0=pRef.copy()
  p01=pRef.copy()
# select zone around line  center
  #x1=p0[0][1]-zone;   x2=p0[0][1]+zone
  x1 = xmin_fit ; x2 = xmax_fit
  x=numpy.arange(data1.shape[1])[x1:x2]  #  x-axis array in pix
  spe=data1[fiber,x1:x2]  #read spectrum in the line range 

  ll=numpy.where(spe == max(spe) )
  p0[0][1]= ll[0][0]+x1   

  #fitfunc = lambda p0, x: p0[0]*scipy.exp(-(x-p0[1])**2/(2.0*p0[2]**2))
  fitfunc = lambda p0, x: p0[3] + p0[0]*scipy.exp(-(x-p0[1])**2/(2.0*p0[2]**2))
  errfunc = lambda p, x, y: fitfunc(p,x)-y
  #-p1, success= scipy.optimize.leastsq(errfunc, p0.copy()[0],args=(x,spe))
  p1, success= scipy.optimize.leastsq(errfunc, p0.copy()[0],args=(x,spe))
  
  ref= fitfunc(pRef[0], x)
  fit= fitfunc(p1, x)
  return success, p1, x, spe, ref, fit



  
# -----------------
def checkOneFile(mjd,fibers,fExp,sHeader, outfile):
    for fiber in fibers:      
         ss="%s %3i" %  (sHeader, fiber)  # row
         for chip in ["a","b","c"]: 

            ff= getFileName(path, mjd, chip, fExp)
            #hdulist=pyfits.open(ff,'readonly')
            #data1=hdulist[1].data
            #hdulist.close()
            hdu1 = fits.open(ff)
            data1 = hdu1[1].data


            pp="p0%s" % chip
            pRef=eval(pp)                
            (succ,p1,x,spe,ref,fit)=OneFileFitting(data1, fiber, pRef)                 
            sProf="%5.2f %5.2f %5.2f" % \
                  (p1[0]/pRef[0][0],  # intensity
                   p1[1] - pRef[0][1],  # profile center
                   p1[2] - pRef[0][2] ) # width
            ss="%s  %s" % (ss, sProf)      
                   
   #      print "%s " % ss
         myPrint(ss, outfile)
    return True
  
#-----------------------------------------  
def checkOneMjd(mjd, dither, fibers, outfile):
  mask=getFileName(path, mjd, "a")
  files = sorted(glob.glob(mask))
  if len(files)==0:  
       return False
  for ff in files:  
     #hdulist=pyfits.open(ff,'readonly')
     #hdr = hdulist[0].header
     #hdulist.close()
     hdu1 = fits.open(ff)
     data1 = hdu1[0].data

     dth=float(hdr['DITHPIX'])
#     q0= sdth(dth) in dither
     q1= hdr.get('IMAGETYP')=="ArcLamp"
     q2= hdr.get('LAMPTHAR')==1
#     if q0 and q1 and q2:
     if q1 and q2:
         fExp=ff[35:43] 
         ss="%5s %8s %s" %  (mjd,fExp,sdth(dth))                  
         checkOneFile(mjd,fibers,fExp,ss, outfile)
  return True
     
def myPrint(ss, outfile): 
  print (ss)
  outfile.write(ss+'\n')
  outfile.flush()

#--------------------------  
def main(argv=None):
  if argv is None: argv = sys.argv[1:]

# current mjd
#  TAI_UTC =34; sjd1=(time.time() + TAI_UTC) / 86400.0 + 40587.3;  sjd= int (sjd1)

  desc = 'apogee arc Thar check for mjd range'
  parser = argparse.ArgumentParser(description=desc)
  parser.add_argument('-m1', '--mjd1', help='start mjd, default current', default=curSjd()-1,  type=int)
  parser.add_argument('-m2', '--mjd2', default=curSjd(), help='end mjd, default is the same as mjd1', type=int)
  parser.add_argument('-r', '--ref',  action="store_true",  help='references for  lines')
  parser.add_argument('-c', '--cfull',  action="store_true",  help='do three fibers [30,150,270], default one [150]')
  parser.add_argument('-f', '--fiber',  default=150,  help='select one fiber, default [150]', type=int)
  parser.add_argument('-A', '--ditherA', action="store_true",  help='dither A, default both A and B')
  parser.add_argument('-B', '--ditherB', action="store_true", help='dither B, default both A and B')
  parser.add_argument('-o', '--outfile',  help='output file name')
  parser.add_argument('-s', '--sortList',  action="store_true", help='to sort the listing')
  
  args = parser.parse_args()    
  
  mjd1=args.mjd1
  mjd2=args.mjd2
#  if mjd2==None:  mjd2=curSjd()
  mjds=sorted(range(mjd1, mjd2+1))

  fibers=[args.fiber]
  if args.cfull: 
      fibers=[30, 150, 270]

  dither=[]
  if args.ditherA: dither.append("A")
  if args.ditherB: dither.append("B")
  if len(dither)==0: 
       dither=["A","B"]

  outfileName=args.outfile
  if outfileName==None:
       outfileName="apThar_%5i.outfile" % (mjd2)
  outfile = open(outfileName,'w')
    
  title="# ./apogeeThar.py -m1 %s  -m2 %s" % (mjd1, mjd2)
  separator="#%s" % ("-"*79)  
#  header1 = "# mjd  nExp  Dth fiber"+" "*8+"a"+" "*18+"b"+" "*18+"c"
#  spHed="I/Io  X-Xo  W-Wo"
#  header2 = "#%s" % (" "*22+spHed+" "*3+spHed+" "*3+spHed)
#  ss="%s\n%s\n%s\n%s" %  (separator, header1, header2, separator)
  def sBl(a):  return "(I%s    X%s    W%s)" % (a,a,a) 
  header="# mjd  nExp  Dth fiber %s   %s   %s" % (sBl("a"),sBl("b"),sBl("c"),) 
  myPrint("%s\n%s\n%s\n%s" %  (title, separator, header, separator), outfile)
  
  if args.sortList:
    for dd in dither:
       for fbr in fibers: 
          ff1=[fbr]
          for mjd in mjds: 
              checkOneMjd(mjd, dd, ff1, outfile)
       myPrint(separator, outfile)        
  else:   
     for mjd in  mjds:
         checkOneMjd(mjd, dither, fibers, outfile)
         myPrint(separator, outfile)
      
  if  args.ref:
     for line in ['a','b','c']:
        pp="p0%s" % line;   p0=eval(pp)     
        myPrint("# ref-%s:  [%6i,  %7.2f,  %4.2f]"% (line, p0[0][0], p0[0][1],p0[0][2]), outfile)

  print ("" )


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""
help.py

This is a file to pull the help for all programs in bin. Each script must be
 added individually, some scripts use argparse, so -h is called, others do not,
 and so their __doc__ is printed. Based on spHelp written by Elena.

2020-06-08  DG  Init, based on spHelp
2020-08-17  DG  Added XMID and WAVEMID
2020-10-18  DG  Added m4l
2022-01-09  DG Simplified the whole thing to text
"""

__version__ = '3.2.0'


doc="""
ObserverTools

Most of these tools have their own --help explanations that are more verbose

Scripts:
ads9.py: Shortcut for ds9_live.py -a
ap_test.py: Tests apogee dome flats (in progress)
aptest: Alias for ap_test.py
az_fiducial_error.sh: Checks mcp logs for fiducial crossings
boss_shap1sum.py: Hashes boss directories
ds9_live.py: Runs a ds9 window for any standard image directory
eval_pointint.py: Analyzes ecam pointing datasets
fsc_coord_conver.py: Converts tables of fsc coordinates to sky offsets
get_dust.py: Gets today's dust counts
getDust.py: Alias for get_dust.py
help.py: This script
influx_fetch.py: Queries InfluxDB, requires an influxdb key
list_ap: Prints apogee images, I recommend sloan_log.py -a instead
m4l: Alias for m4l.py
m4l.py: Gets Mirror numbers
mjd: Alias for sjd.py
plot_mcp_fiducials.py: Plots mcp fiducial crossings
sjd.py: Gets current sjd
sloan_log.py: A large and flexible logging tool including apogee, boss,
 log support, and much more.
sossy.py: Parse SOS outputs
spds9.py: Shortcut for ds9_live.py -b
spHelp: Alias for help.py
spVersion: Alias for versions.py
telescope_status.py: Creates telescope status section for night log
telStatus: Alias for telescope_status.py
time_track.py: A time tracking tool to anlayze monthly time usage
tpm_feed.py: Creates a feed/plot of TPM output
tpm_fetch.py: Queries TPM archive files without InfluxDB
versions.py: Prints various software versions and disk usage
wave_mid.py: Computes WAVEMID offsets
WAVEMID: Alias for wave_mid.py
x_mid.py: Computes XMID offsets
XMID: Alias for x_mid.py
"""
print(doc)



#!/usr/bin/env python3
import unittest
import subprocess as sub
from bin import versions


class TestSPVersion(unittest.TestCase):
    def test_no_args(self):
        output = sub.Popen(versions.__file__, stderr=sub.PIPE,
                          stdout=sub.PIPE)
        for line in output.stderr.read().decode('utf-8').splitlines():
            print(line)
            self.assertNotIn('not found', line,
                             msg='Package manager or package not found')
        # self.assertEqual(output.returncode, 0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

from pathlib import Path

data_path = Path("/data/")  # For hub/normal computers
if not data_path.exists():
    data_path = Path("/data_hub/")  # For obs1 and obs2
    if not data_path.exists():
        data_path = Path("/Volumes/data/")  # For sdss-obs3
        if not data_path.exists(): 
            data_path = Path.home() / "data/"  # For testing, usually on a Mac
            if not data_path.exists():
                raise FileNotFoundError("Cannot find the SDSS data directory")
else:
    ap_utr: Path = data_path / "apogee/utr_cdr"
    try:
        ap_utr.exists()
    except PermissionError:  # This happens on obs1/2 with /data_hub, even
        # though /data exists, it cannot be read.
        data_path = Path("/data_hub/")

ap_utr: Path = data_path / "apogee/utr_cdr"
ap_archive: Path = data_path / "apogee/archive"
ap_qr: Path = data_path / "apogee/quickred"
boss: Path = data_path / "spectro"
dos: Path = data_path / "manga/dos"
sos: Path = data_path / "boss/sos"
ecam: Path = data_path / "ecam"
gcam: Path = data_path / "gcam"
fsc: Path = data_path / "fsc"
fvc: Path = data_path / "fcam"
mcp_logs: Path = data_path / "logs/mcp"



import unittest
from bin import eval_pointing

class TestEvalPointing(unittest.TestCase):
    def setUp(self) -> None:
        class Args:
            pass
        self.args = Args()
        self.args.mjd = 59464
        self.args.verbose = False
        self.args.window = "285-364"
        self.args.master_field = 285
        self.args.plot = True
        self.args.plot_file = ""
        self.args.json = ""
        self.args.file = ""
        self.args.threshold = 0
        self.args.json = False
    
    def test_default_night(self):
        eval_pointing.main(self.args)



#!/usr/bin/env python3
"""
This is a flexible tool used to query the SDSS-V InfluxDB. It is usually very
fast, especially compared to the old EPICS, and it is meant to be a tool
that can be imported as well as run through a CLI

Author: Dylan Gatlin
"""
import os
import platform
import argparse

import subprocess as sub

from pathlib import Path
from influxdb_client import InfluxDBClient
from astropy.time import Time

from bin import sjd

__version__ = "3.0.0"
__author__ = "Dylan Gatlin"


def ping(host):
    """
    Returns True if host (str) responds to a ping request.
     Remember that a host may not respond to a ping (ICMP) request even if the
     host name is valid.
    """

    # Option for the number of packets as a function of
    param = '-n' if platform.system().lower() == 'windows' else '-c'

    # Building the command. Ex: "ping -c 1 google.com"
    command = ['ping', param, '1', "-i", "0.2", host]
    x = sub.run(command, stdout=sub.PIPE)
    return x.returncode == 0


def get_key():
    """Finds a file called .influx.key that has 3 lines, a user id, an org id,
    and a token, each on its own line. The file can be in the current directory
    or the home directory
    TODO: Put an influx.key somewhere that all observers can use it.
    """
    if "INFLUXDB_V2_TOKEN" in os.environ.keys():
        user_id = os.environ["INFLUXDB_V2_USER"]
        org_id = os.environ["INFLUXDB_V2_ORG"]
        token = os.environ["INFLUXDB_V2_TOKEN"]
        return user_id, org_id, token
        
    key_path = Path.home() / ".influx.key"
    if not key_path.exists():
        key_path = Path(".") / ".influx.key"
    elif not key_path.exists():
        raise KeyError("Influx Key not found, cannot send a query without one")
    with key_path.open('r') as fil:
        user_id = fil.readline().rstrip('\n')
        org_id = fil.readline().rstrip('\n')
        token = fil.readline().rstrip('\n')
    return user_id, org_id, token


def get_client(org_id, token):
    if ping("10.25.1.221"):
        client = InfluxDBClient(url="http://10.25.1.221:8086", token=token,
                                org=org_id, timeout=20000)
    else:
        # print("Did not reach 10.25.1.221")
        client = InfluxDBClient(url="http://localhost:8086", token=token,
                                org=org_id, timeout=20000)
    return client.query_api()




def query(flux_script, start, end, interval="1s"):
    user, org, token = get_key()
    client = get_client(org_id=org, token=token)
    query = flux_script
    query = query.replace("v.timeRangeStart", f"{start.isot}Z")
    query = query.replace("v.timeRangeStop", f"{end.isot}Z")
    query = query.replace("v.windowPeriod", interval)
    result = client.query(query=query, org=org)
    return result
    

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--mjd", type=float, help="SJD of observations")
    parser.add_argument("--t1", "--start-time", dest="start_time", help="Start"
        " time for astropy.time to parse, preferable isot")
    parser.add_argument("--t2", "--end-time", dest="end_time", help="Start time"
        " for astropy.time to parse, preferable isot")
    parser.add_argument("-f", "--file", nargs='+', help="A file path of a .flux"
        " influxdb query file")
    parser.add_argument("-i", "--interval", default="1m", help="Time interval")
    parser.add_argument("-v", "--verbose", action="store_true",
        help="Verbose debugging")

    args = parser.parse_args()

    if (not args.mjd ) and (not args.start_time and not args.end_time):
        args.start_time = Time(sjd.sjd(), format="mjd") - 0.3
        args.end_time = Time.now()
    elif args.mjd:
        args.start_time = Time(args.mjd - 1, format="mjd")
        args.end_time = Time(args.mjd, format="mjd")
    elif args.start_time:
        args.start_time = Time(args.start_time)
        args.end_time = Time(args.end_time)
    return args


def main(args=None):
    if args is None:
        args = parse_args()
    user_id, org_id, token = get_key()
    client = get_client(org_id, token)
    if args.verbose:
        print(f"Querying from {args.start_time.isot} to {args.end_time.isot}")
    if args.file:
        for f_path in args.file:
            f_path = Path(f_path)
            if f_path.exists():
                with f_path.open('r') as fil:
                    query = fil.read()
        query = query.replace("v.timeRangeStart", f"{args.start_time.isot}Z")
        query = query.replace("v.timeRangeStop", f"{args.end_time.isot}Z")
        query = query.replace("v.windowPeriod", args.interval)
        if args.verbose:
            print(query)
        query_result = client.query(org=org_id, query=query)
        print(f"{'Time':23} {'Measurement':15} {'Field':15} {'Value':}")
        print("=" * 80)
        for res in query_result:
            for val in res:
                if "_measurement" in val.values.keys():
                    print(f"{Time(val.get_time()).isot:23}"
                          f" {val.get_measurement():15}"
                          f" {val.get_field():15}"
                          f" {val.get_value()}")
                else:
                    print(f"{Time(val.get_time()).isot:23}",
                          val
                          )
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""A tool to convert from the FSC coordinates of r, theta to standard RA, Dec
 for tcc
"""
import numpy as np
from argparse import ArgumentParser, ArgumentError

__author__ = "Dylan Gatlin"
__version__ = "3.0.0"


def converter(r, theta):
    return r * np.sin(np.deg2rad(theta)), r * np.cos(np.deg2rad(theta))


def arg_parse():
    parser = ArgumentParser(description="A tool for converting coordinates to"
            " from FSC's r, theta to RA and Dec offsets for tcc. If given with"
            " arguments, give them in pairs of space-separated r and thetas."
            " You can put multiple, as long as they're in order. If no pairs"
            " are given, then a default value will be filled which will cover"
            " the entire focal plane and is what we have done in the past.")
    parser.add_argument("rthetapairs",
            help="r and theta, separated by spaces, as many times as you like."
                 " in sky degrees (1.5deg = 333.3mm)", nargs='*', type=float)
    parser.add_argument("--tcc-offset", action="store_true", help="Print tcc"
            " offset commands instead of a table")
    return parser.parse_args()


def main(args=None):
    if args is None:
        args = arg_parse()
    if len(args.rthetapairs) % 2 != 0:
        raise ArgumentError("rthetapairs must be an even number of args!")
    if len(args.rthetapairs) == 0:
        for r in np.array([1.5, 1.125, 0.75, 0.375]):
            for th in np.arange(180, -180, -60):
                args.rthetapairs.append(r)
                args.rthetapairs.append(th)
    if not args.tcc_offset:
        print(f"{'R':10} {'Theta':10} {'RA':10} {'Dec':10}")
    for i in range(len(args.rthetapairs) // 2):
        r = args.rthetapairs[i * 2]
        theta = args.rthetapairs[i * 2 + 1]
        ra, dec = converter(r, theta)
        if not args.tcc_offset:
            print(f"{r:9.5f}\N{DEGREE SIGN} {theta:9.1f}\N{DEGREE SIGN}"
            f"{ra:9.5f}\N{DEGREE SIGN} {dec:9.5f}\N{DEGREE SIGN}")
        else:
            print(f"tcc offset arc/pabs /computed {ra:.5f}, {dec:.5f}")
    return 0


if __name__ == "__main__":
    main()




#!/usr/bin/env python3
"""In order to evaluate the quality of a pointing model, this script checks
a series of images from the engineering camera and evaluates the pointing
accuracy"""

__author__ = "Dylan Gatlin"

import tqdm
import json
import copy
import numpy as np
import matplotlib as mpl
# mpl.use("GTK3Agg")
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from pathlib import Path
from argparse import ArgumentParser
from astropy.io import fits
from astropy.time import Time
from astropy.stats import sigma_clipped_stats
from photutils.detection import DAOStarFinder
from photutils.aperture import CircularAperture

from bin import sjd
from sdssobstools import sdss_paths

class ECamData:
    def __init__(self, master_img: int = None):
        self.times = []
        self.img_nums = []
        self.ras = []
        self.decs = []
        self.azs = []
        self.alts = []
        self.rots = []
        self.stars = []
        self.seeings = []
        self.master_img = master_img

    def sort(self):
        self.times = Time(self.times)
        sorter = np.argsort(self.times)
        self.img_nums = np.array(self.img_nums)[sorter]
        self.ras = np.array(self.ras)[sorter]
        self.decs = np.array(self.decs)[sorter]
        self.azs = np.array(self.azs)[sorter]
        self.alts = np.array(self.alts)[sorter]
        self.rots = np.array(self.rots)[sorter]
        tmp_stars = copy.deepcopy(self.stars)
        # Stars is an object array, hard to work with
        # self.stars = np.array(self.stars)[sorter]
        self.stars = [tmp_stars[i] for i in sorter]
        self.seeings = np.array(self.seeings)[sorter]
        self.master_i = np.where(self.master_img == self.img_nums)[0][0]
        if self.master_img is not None:
            self.brightest_i = np.where(self.stars[self.master_i]["flux"]
                                        == self.stars[self.master_i]["flux"].max())[0][0]

    def build_set(self):
        avg_stars = np.mean([len(s) for s in self.stars])
        master_n_stars = len(self.stars[self.master_i])
        if round(avg_stars) != master_n_stars:
            print(f"The number of stars in the master image"
                  f" ({master_n_stars}) doesn't match the"
                  f" average number of stars ({avg_stars:.1f})")

        self.coord_pairs = np.empty(
            (len(self.stars), master_n_stars, 2), dtype=float)
        self.coord_pairs[:] = np.nan  # Guilty until proven innocent
        for j, stars in enumerate(self.stars):
            for k, ref in enumerate(self.stars[self.master_i]):
                for s in stars:
                    if ((np.abs(s["xcentroid"] - ref["xcentroid"]) < 15)
                            and np.abs(s["ycentroid"] - ref["ycentroid"]) < 15):
                        self.coord_pairs[j, k] = np.array([s["xcentroid"],
                                                           s["ycentroid"]])
                        continue

        return self.coord_pairs

    def to_json(self, outfile):
        output = {}
        output["Time"] = [t.iso for t in self.times]
        output["ImgID"] = np.ndarray.tolist(self.img_nums)
        output["RA"] = np.ndarray.tolist(self.ras)
        output["Dec"] = np.ndarray.tolist(self.decs)
        output["Alt"] = np.ndarray.tolist(self.alts)
        output["Az"] = np.ndarray.tolist(self.azs)
        output["Rot"] = np.ndarray.tolist(self.rots)
        output["Coords"] = np.ndarray.tolist(self.coord_pairs)
        output["Brightest"] = int(self.brightest_i)
        if self.master_img:
            output["MasterID"] = int(self.master_img)
            output["MasterInd"] = int(self.master_i)
        json.dump(output, outfile.open('w'), indent=4)
        return

    def __iter__(self):
        return zip(self.times, self.img_nums, self.ras, self.decs, self.azs,
                   self.alts, self.rots, self.coord_pairs)


ecam_mask = np.zeros((512, 524), dtype=bool)
ecam_mask[:6, :] = True
ecam_mask[-5:, :] = True
ecam_mask[:, :6] = True
ecam_mask[:, -10:] = True  # This side has lots of bad pixels


def analyze_ecam(img_path, data_class, args):
    img = fits.open(img_path.as_posix())
    if img[0].header["IMAGETYP"] != "object":
        return
    data_class.times.append(img[0].header["DATE-OBS"][:19])
    data_class.img_nums.append(int(img_path.name.split('-')[-1].split('.')[0]))
    data_class.ras.append(img[0].header["RA"])
    data_class.decs.append(img[0].header["DEC"])
    data_class.azs.append(img[0].header["AZ"])
    data_class.alts.append(img[0].header["ALT"])
    data_class.rots.append(img[0].header["ROTPOS"])
    data_class.seeings.append(img[0].header["SEEING"])  # Usually 0
    mean, median, std = sigma_clipped_stats(img[0].data)
    if not args.threshold:
        daofind = DAOStarFinder(fwhm=5, threshold=mean + std * 5)
    else:
        daofind = DAOStarFinder(fwhm=5, threshold=args.threshold)
    sources = daofind(img[0].data, mask=ecam_mask)
    if sources is None:
        if args.verbose:
            print(f"    Skipping {img_path.name}")
        data_class.stars.append(sources)  # Prevents operations on nonetype
        # while maintaining data_class' list lengths

        return
    sources = sources[sources["flux"] > 1.5]  # Gets rid of the faint targets
    data_class.stars.append(sources)
    if args.verbose:
        print(f"File: {img_path.name} NSource: {len(sources)},"
              f" Mean: {mean:.1f}, Std: {std:.1f}")
    return img


def plot_img(fits_obj, sources, args):
    mean, median, std = sigma_clipped_stats(fits_obj[0].data)
    positions = np.transpose((sources["xcentroid"], sources["ycentroid"]))
    apertures = CircularAperture(positions, r=10)
    fig, ax = plt.subplots(1, 1)
    ax.imshow(fits_obj[0].data, cmap="Greys", origin="lower",
              vmin=mean+std, vmax=mean+std*10, interpolation="nearest")
    apertures.plot(axes=ax, color="blue", lw=1.5, alpha=0.5)
    mask = Rectangle((5, 5), width=524-15, height=512-10, alpha=1, fill=False,
                     linestyle="--", color="red", label="Mask")
    ax.add_patch(mask,)
    if args.plot_file:
        fig.savefig(args.plot_file)
    else:
        plt.show()


def print_fit_table(data_class, master_i: int = None):
    print(f"{'Time':20}{'Image':6}{'RA':6}{'Dec':6}{'Alt':6}{'Az':6}{'Rot':6}"
          f"{'Offset':6}")
    print('=' * 80)
    for t, n, r, d, al, az, rot, s in data_class:
        try:
            brightest = (data_class.stars[master_i]["flux"]
                         == data_class.stars[master_i]["flux"].max())
        except (ValueError, TypeError):  # For some empty star lists
            continue

        if master_i is None:
            offset = np.sqrt((s[brightest, 0] - 524 // 2)**2
                             + (s[brightest, 1] - 512 // 2)**2)[0]
        else:
            offset = np.sqrt(
                (s[brightest, 0]
                    - data_class.coord_pairs[master_i, brightest, 0])**2
                + (s[brightest, 1]
                    - data_class.coord_pairs[master_i, brightest, 1])**2)[0]

        print(f"{t.iso[:19]:19}{n:6.0f}{r:6.1f}{d:6.1f}{al:6.1f}{az:6.1f}"
              f"{rot:6.1f}{offset:6.1f}")


def parse_args():
    parser = ArgumentParser("A tool to evaluate SDSS' pointing and tracking."
                            " If no arguments are given, it will check through"
                            " an entire night's worth of data and print a table"
                            " of offsets from the center for the brightest star."
                            " If a window is specified, you'll get results"
                            " split by the specified windows. If a file is"
                            " specified, you can check a single image.")
    parser.add_argument("-m", "--mjd", default=sjd.sjd())
    parser.add_argument("-t", "--threshold", type=int,
                        help="A lower limit on the peak"
                             " flux of a target. By default, it is 3 sigma"
                             " above the mean, but an integer could be provided"
                             " instead.")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Verbose debugging output")
    parser.add_argument("-f", "--file", nargs="+",
                        help="Specific files to process")
    parser.add_argument("-w", "--window",
                        help="A dash separated beginning and end of the window."
                             " Always include a --master-field with your window."
                             " Ex: -w 15-50 -s 35")
    parser.add_argument("-s", "--master-field", type=int,
                        help="The master field with all stars you want. Only"
                        " used in conjunction with --window. You need both"
                        " arguments.")
    parser.add_argument("-p", "--plot", action="store_true",
                        help="Show plot to analyze fit quality")
    parser.add_argument("-k", "--plot-file", help="Filename to save plot to."
                        " only works with --plot and --window together. If not"
                        " included, the plot will show up in a new window (or"
                        " maybe not depending on X11).")
    parser.add_argument("-j", "--json", type=str,
                        help="Output a dataset to a json file")
    # TODO Add a way to ignore stars within a field's window
    args = parser.parse_args()
    return args


def main(args=None):
    if args is None:
        args = parse_args()
    if args.verbose:
        print(f"ECamera mask covers {ecam_mask.sum()/ecam_mask.size:.1f}%"
              f" of the total image")
    if args.file:
        ecam = ECamData()
        for i, fil in enumerate(args.file):
            fil_path = Path(fil)
            if fil_path.exists():
                fits_img = analyze_ecam(fil_path, ecam, args)
            else:
                raise FileNotFoundError(
                    f"File {fil_path.absolute()} not found")
            if args.plot:
                plot_img(fits_img, ecam.stars[i], args)
    elif args.window:
        low, high = args.window.split("-")
        try:
            low = int(low)
            high = int(high)
        except ValueError:
            raise ValueError(f"Window range must be filled with ints"
                             f" {args.window}")
        ecam = ECamData(args.master_field)
        for j in tqdm.tqdm(list(range(low, high+1))):
            ecam_path = (
                sdss_paths.ecam / f"{args.mjd}/proc-gimg-{j:04.0f}.fits.gz")
            analyze_ecam(ecam_path, ecam, args)
        ecam.sort()
        coord_pairs = ecam.build_set()
        if args.json:
            ecam.to_json(Path(args.json))
        if args.plot:
            fig, axs = plt.subplots(2, 2, figsize=(16, 8), sharex=True)
            fig.set_tight_layout("tight")
            # TODO set a suptitle of the field position
            labels = ["x", "y"]
            brightest = (ecam.stars[ecam.master_i]["flux"]
                         == ecam.stars[ecam.master_i]["flux"].max())
            for i, ax in enumerate(axs[0]):
                for j, star in enumerate(coord_pairs[:, :, i].T):
                    ax.plot_date(ecam.times.plot_date[~np.isnan(star)],
                                 (star[~np.isnan(star)] -
                                  star[ecam.master_i]) * 0.428,
                                 fmt="-", alpha=0.3)
                ax.plot_date(ecam.times.plot_date,
                             np.nanmean(coord_pairs[:, :, i]
                                        - coord_pairs[ecam.master_i, :, i],
                                        axis=1) * 0.428,
                             fmt="-", label="Mean")
                ax.plot_date(ecam.times.plot_date,
                             (coord_pairs[:, brightest, i]
                              - coord_pairs[ecam.master_i, brightest, i]) * 0.428,
                             fmt="-", label="Brightest")
                ax.set_ylabel(f"{labels[i]} Axis Drift (arcsec)")
                ax.legend()
            axs[1, 0].plot_date(ecam.times.plot_date[~np.isnan(ecam.coord_pairs[:, brightest, 0]).T[0]],
                                np.sqrt((ecam.coord_pairs[~np.isnan(ecam.coord_pairs[:, brightest, 0].T[0]), brightest, 0]
                                         - ecam.coord_pairs[ecam.master_i, brightest, 0])**2
                                + (ecam.coord_pairs[~np.isnan(ecam.coord_pairs[:, brightest, 0]).T[0], brightest, 1]
                                   - ecam.coord_pairs[ecam.master_i, brightest, 1])**2) * 0.428,
                                fmt="-", label="Brightest")
            axs[1, 0].set_xlabel("Time")
            axs[1, 0].set_ylabel(
                r"Total Drift $\sqrt{\Delta x^2+\Delta y^2}$ (arcsec)")
            axs[1, 0].legend()
            axs[1, 1].set_xlabel("Time")
            if args.plot_file:
                fig.savefig(args.plot_file)
            else:
                plt.show()
        else:
            print_fit_table(ecam, ecam.master_i)

    else:
        ecam = ECamData()
        day_path = (sdss_paths.ecam / f"{args.mjd}").absolute()
        if not day_path.exists():
            raise NotADirectoryError(
                f"Directory {day_path.as_posix()} not found")

        for proc in tqdm.tqdm(list(day_path.glob("proc-gimg-*.fits.gz"))):
            if args.verbose:
                print(proc.absolute())
            analyze_ecam(proc, ecam, args)
        ecam.sort()
        print_fit_table(ecam)

    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Written by Dylan. This script gets the dust counts in a
 night and prints it

Changelog:
2020-06-20  DG  Moving main contents into a function for import functionality
 """
import argparse
from astropy.time import Time
from pathlib import Path
from bin import sjd, influx_fetch

__version__ = '3.3.1'


def parse_args():
    today = sjd.sjd()
    desc = ('Integrates the dust accumulated in a night using TRON and prints'
            ' it. Needs no args for the current date.')
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('-m', '--mjd', help='enter mjd, default is current mjd',
                        default=today, type=int)
    parser.add_argument('-v', '--verbose', action="store_true",
                        help='print incremental dust data')
    args = parser.parse_args()
    
    if (not args.mjd ) and (not args.start_time and not args.end_time):
        args.start_time = Time(sjd.sjd(), format="mjd")
        args.end_time = Time.now()
    elif args.mjd:
        args.start_time = Time(args.mjd, format="mjd")
        args.end_time = Time(args.mjd + 1, format="mjd")
    return args


def get_dust(start_time, end_time, verbose):
    q_path = Path(__file__).parent.parent / "flux/dust.flux"
    if not q_path.exists():
        raise FileNotFoundError(f"Couldn't find Flux query {q_path.absolute()}")
    query = q_path.open('r').read()
    if verbose:
        print(f"Start: {start_time.isot}, End: {end_time.isot}")
    result = influx_fetch.query(query, start_time, end_time)
    if len(result) == 0:
        dust_sum = 0
    else:
        # If I ever stop trusting cumulateiveSum() to do the summation for me,
        # this routine is a decent substitute.
        # times = []
        # vals = []
        # for row in result[0].records:
            # times.append(row.get_time())
            # vals.append(row.get_value())
        # times = Time(times).mjd
        # vals = np.array(vals)
        # dust_sum = np.sum(np.gradient(times) * 24 * vals )
        dust_sum = result[0].records[-1].get_value()
        
    if len(result) == 0:
        return dust_sum
    if verbose:
        for row in result[0].records:
            print(row.get_time(), row.get_value())
    return dust_sum

def main(args=None):
    if args is None:
        args = parse_args()
    dust_sum = get_dust(args.start_time, args.end_time, args.verbose) 
    print("Integrated Dust Counts: ~{:<.0f} dust-hrs".format(
          dust_sum - dust_sum % 100))



if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""sloan_log.py

A script to automate the bulk of logging and to replace various tools
 like log function, log support, list_ap, list_m, and more. This code is
 entirely dependent on raw images, and their headers, unlike Log Function
 which is dependent on callbacks it catches while open only and is subject to
 crashes. It's inspired by some outputs from time tracking, but it avoids the
 sdss python module and platedb

The intent of this is to create a more future-proof tool for logging through
Sloan V

In order to run it locally, you will need to either have access to /data, or
 fake it with a single night. You'll need a date from /data/spectro and
 /data/apogee/archive. You'll also need setup local forwarding for InfluxDB
"""
import argparse
import sys
import textwrap
import warnings
import fitsio

import numpy as np
try:
    from bin import m4l, telescope_status, get_dust
except ImportError as e:
    raise ImportError('Please add ObserverTools/bin to your PYTHONPATH:'
        '\n    {}'.format(e))
from pathlib import Path
from tqdm import tqdm
from astropy.time import Time

from sdssobstools import apogee_data, log_support, boss_data, sdss_paths
    
from bin import sjd

if sys.version_info.major < 3:
    raise Exception('Interpretter must be python 3.5 or newer')

# For astropy
warnings.filterwarnings('ignore', category=UserWarning, append=True)
# For numpy boolean arrays
warnings.filterwarnings('ignore', category=FutureWarning, append=True)

__version__ = '3.7.5'

ap_dir = sdss_paths.ap_archive
b_dir = sdss_paths.boss

class Logging:
    """
    A tool to produce a ton of various outputs used for logging. This tool uses
    the images in the /data directory and epics to build a nightlog that doesn't
    require STUI output. It is best run as follow:

    log = Logging(ap_images, b_images, args)
    log.parse_images()
    log.sort()
    log.count_dithers()
    log.p_summary()
    log.p_data()
    log.p_boss()
    log.p_apogee()
    log.log_support()

    or via command line as

    ./sloan_log.py -pt

    or for a previous date

    ./sloan_log.py -pm 59011

    """

    def __init__(self, ap_images, m_images, args):
        self.ap_images = ap_images
        self.b_images = m_images
        self.args = args
        # Dictionary keys that begin with c are of len(carts_in_a_night),
        # keys that begin with i are of len(images_in_a_night), keys that begin
        # with d are of len(dome_flats) (which is not always the same as carts,
        # keys that begin with a are of len(apogee_arcs), which is usually 4 in
        # a full night (morning and evening cals), keys that begin with h are of
        # len(hartmanns), which is usually a few longer than len(carts). Each
        # first letter is used to choose which sorting key to use. There must be
        # a Time key for each new letter, and a new sorter argument must be
        # added to self.sort. All of these dictionary items begin as lists,
        # and are converted to np.arrays or astropy.time.Times in self.sort.
        self.data = {'dDesign': [], 'dTime': [], 'dConfig': [], 'cLead': []}
        self.ap_data = {'dDesign': [], 'dTime': [],
                        'iTime': [], 'iID': [],
                        'iSeeing': [], 'iDetector': [], 'iDither': [],
                        'iNRead': [], 'iEType': [], 'iDesign': [], 
                        "iDesign": [], "iConfig": [],
                        'fDesign': [], 'fTime': [], 'fMissing': [], 'fFaint': [],
                        'fNMissing': [], 'fNFaint': [], 'fRatio': [], 'aTime': [],
                        'aOffset': [], 'aID': [], 'aLamp': [], 'oTime': [],
                        'oOffset': [], 'oDither': []}
        self.b_data = {'dDesign': [], 'dTime': [],
                       'iTime': [], 'iID': [],
                       'iDetector': [], 'iDither': [],
                       'iEType': [], 'idt': [], 'iDesign': [], 'iHart': [],
                       'iDesign': [], "iConfig": [], 'hHart': [], 'hTime': []}
        # These values are not known from the header and must be created
        # after self.sort. N for number, AP or B for APOGEE or BOSS, and NSE
        # for BOSS dithers, and AB for APOGEE dithers, dt for boss exposure
        # time, since some boss carts use shorter exposures. All these are
        # combined to fill Summary in self.count_dithers
        self.design_data = {'cNAPA': [], 'cNAPB': [], 'cNBN': [], 'cNBS': [],
                          'cNBE': [], 'cNBC': [], 'cBdt': [], 'cNB': [],
                          'cAPSummary': [],
                          'cBSummary': []}
        self.test_procs = []
        master_path = (Path(apogee_data.__file__).absolute().parent.parent
                           / "dat/master_dome_flat.fits.gz")
        if not master_path.exists():
            master_path = (Path(apogee_data.__file__).absolute(
                ).parent.parent.parent / "dat/master_dome_flat.fits.gz")
        master_data = fitsio.read(master_path.as_posix())
        self.ap_master = np.median(master_data[:, 550:910], axis=1)
        self.morning_filter = None
        self.ap_image = None

    def ap_test(self, img):
        """Calls aptest on hub, this could certainly be replaced in the near
        future.
        """
        # This is from ap_test
        self.args.plot = False
        missing, faint, flux_ratio = img.ap_test((550, 910), self.ap_master)
        if self.args.verbose:
            print('Exposure {}'.format(img.exp_id))
            print(missing, faint)
        n_missing = 0
        n_faint = 0
        for miss in missing:
            if isinstance(miss, bytes) or isinstance(miss, str):
                if 'bundle' in miss:
                    n_missing += 30
                else:
                    n_missing += abs(eval(miss))
            else:
                n_missing += 1
        for fain in faint:
            if isinstance(fain, bytes) or isinstance(fain, str):
                if 'bundle' in fain:
                    n_faint += 30
                else:
                    n_faint += np.abs(eval(fain))
            else:
                n_faint += 1
        # return (n_missing, n_faint, missing, faint, img.cart_id,
        #         img.isot)
        self.ap_data['fNMissing'].append(n_missing)
        self.ap_data['fNFaint'].append(n_faint)
        self.ap_data['fMissing'].append(missing)
        self.ap_data['fFaint'].append(faint)
        self.ap_data['fRatio'].append(flux_ratio)
        self.ap_data['fDesign'].append(img.design_id)
        self.ap_data['fTime'].append(img.isot)

    def parse_images(self):
        """Goes through every image in ap_images and m_images to put them in
        dictionaries."""
        if self.args.apogee:
            print('Reading APOGEE Data ({})'.format(len(self.ap_images)))
            img = None
            for image in tqdm(self.ap_images):
                if self.args.verbose:
                    print(image.name)
                img = apogee_data.APOGEERaw(image, self.args, 1)
                # img.parse_layer(1)
                if img.lead is None:  # If the first exposure is still
                    # writing, plate_id will be empty and without this if,
                    # it would fail. With this if, it will skip the plate
                    print(f"Skipping {image}")
                    continue
                if (img.exp_type == 'Domeflat') and ('-a-' in img.file.name):
                    self.ap_test(img)
                    self.test_procs.append(img.design_id)
                elif ('Arc' in img.exp_type) and ('-a-' in img.file.name):
                    self.ap_data['aTime'].append(img.isot)
                    self.ap_data['aID'].append(img.exp_id)
                    if 'ThAr' in img.exp_type:
                        self.ap_data['aOffset'].append(
                            img.compute_offset((60, 70), 1105, 40, 1.27))
                        self.ap_data['aLamp'].append('ThAr')
                    elif 'UNe' in img.exp_type:
                        self.ap_data['aOffset'].append(
                            img.compute_offset((60, 70), 1180, 30, 3))
                        self.ap_data['aLamp'].append('UNe')
                    else:
                        print("Couldn't parse the arc image: {} with exposure"
                              " type {}".format(img.file, img.exp_type))
                        self.ap_data["aTime"].pop()
                        self.ap_data["aID"].pop()
                elif ('Object' in img.exp_type) and ('-a-' in img.file.name):
                    # TODO check an object image for a good FWHM (last
                    # input)
                    self.ap_data['oTime'].append(img.isot)
                    self.ap_data['oOffset'].append(
                        img.compute_offset((30, 35), 1090, 40, 2))
                    self.ap_data['oDither'].append(img.dither)

                if img.design_id not in self.data['dDesign']:
                    self.data['dConfig'].append(img.plate_id)
                    self.data['dDesign'].append(img.design_id)
                    self.data['dTime'].append(img.isot)
                    try:
                        self.data['cLead'].append(img.lead)
                    except AttributeError:
                        self.data["cLead"].append("")
                else:
                    i = self.data['dDesign'].index(img.design_id)
                    if img.isot < self.data['dTime'][i]:
                        self.data['dTime'].pop(i)
                        self.data['dTime'].insert(i, img.isot)
                if img.design_id not in self.ap_data['dDesign']:
                    self.ap_data['dDesign'].append(img.design_id)
                    self.ap_data['dTime'].append(img.isot)
                else:
                    i = self.ap_data['dDesign'].index(img.design_id)
                    if img.isot < self.ap_data['dTime'][i]:
                        self.ap_data['dTime'].pop(i)
                        self.ap_data['dTime'].insert(i, img.isot)
                detectors = []
                arch_dir = sdss_paths.ap_archive / f"{self.args.sjd}/"
                # red_dir = Path('/data/apogee/quickred/{}/'.format(
                #     self.args.sjd))
                # This used to see if quickred processed, but others preferred
                # to see if the archive image was written
                arch_name = 'apR-a-{}.apz'.format(img.exp_id)
                if (arch_dir / arch_name).exists():
                    detectors.append('a')
                # elif (red_dir / arch_name.replace('1D', '2D')).exists():
                # detectors.append('2')
                else:
                    detectors.append('x')
                if (arch_dir / arch_name.replace('-a-', '-b-')).exists():
                    detectors.append('b')
                # elif (red_dir / arch_name.replace('-a-', '-b-').replace(
                # '1D', '2D')).exists():
                # detectors.append('2')
                else:
                    detectors.append('x')
                if (arch_dir / arch_name.replace('-a-', '-c-')).exists():
                    detectors.append('c')
                # elif (red_dir / arch_name.replace('-a-', '-c-').replace(
                # '1D', '2D')).exists():
                # detectors.append('2')
                else:
                    detectors.append('x')
                self.ap_data['iTime'].append(img.isot)
                self.ap_data['iID'].append(img.exp_id)
                self.ap_data['iSeeing'].append(img.seeing)
                self.ap_data['iDetector'].append('-'.join(detectors))
                self.ap_data['iDither'].append(img.dither)
                self.ap_data['iNRead'].append(img.n_read)
                self.ap_data['iEType'].append(img.exp_type)
                self.ap_data['iDesign'].append(img.design_id)
                self.ap_data['iConfig'].append(img.config_id)
            if img is not None:
                self.ap_image = img
        if self.args.boss:
            print('Reading BOSS Data ({})'.format(len(self.b_images)))
            for image in tqdm(self.b_images):
                img = boss_data.BOSSRaw(image)
                if img.design_id not in self.data['dDesign']:
                    self.data['dDesign'].append(img.design_id)
                    self.data['dConfig'].append(img.plate_id)
                    try:
                        self.data['cLead'].append(img.lead)
                    except AttributeError:
                        self.data["cLead"].append("")    
                    self.data['dTime'].append(img.isot)
                else:
                    i = self.data['dDesign'].index(img.design_id)
                    if img.isot < self.data['dTime'][i]:
                        self.data['dTime'].pop(i)
                        self.data['dTime'].insert(i, img.isot)
                if img.design_id not in self.b_data['dDesign']:
                    self.b_data['dDesign'].append(img.design_id)
                    self.b_data['dTime'].append(img.isot)
                else:
                    i = self.b_data['dDesign'].index(img.design_id)
                    if img.isot < self.b_data['dTime'][i]:
                        self.b_data['dTime'].pop(i)
                        self.b_data['dTime'].insert(i, img.isot)
                self.b_data['iTime'].append(img.isot)
                self.b_data['iID'].append(img.exp_id)
                # self.b_data['iSeeing'].append(img.seeing)
                self.b_data['iDither'].append(img.dither)
                self.b_data['iEType'].append(img.flavor)
                self.b_data['idt'].append(img.exp_time)
                self.b_data['iDesign'].append(img.design_id)
                self.b_data['iHart'].append(img.hartmann)
                self.b_data['iConfig'].append(img.config_id)

                # if img.hartmann == 'Left' and self.telemetry:
                #     # Note that times are sent in UTC and received in local, yet
                #     # those times are marked as UTC
                #     tstart = Time(img.isot).datetime
                #     tend = (Time(img.isot) + 5 / 24 / 60).datetime
                #     hart = self.telemetry.get([
                #         '25m:hartmann:r1PistonMove',
                #         # '25m:hartmann:r2PistonMove',
                #         '25m:hartmann:b1RingMove',
                #         # '25m:hartmann:b2RingMove',
                #         '25m:hartmann:sp1AverageMove',
                #         # '25m:hartmann:sp2AverageMove',
                #         '25m:hartmann:sp1Residuals:steps',
                #         '25m:hartmann:sp1Residuals:deg',
                #         # '25m:hartmann:sp2Residuals:deg',
                #         '25m:boss:sp1Temp:median',
                #         # '25m:boss:sp2Temp:median',
                #         # '25m:hartmann:sp2Residuals:steps'
                #     ],
                #         start=tstart,
                #         end=tend,
                #         interpolation='raw', scan_archives=False)

                #     self.b_data['hHart'].append(hart)
                #     self.b_data['hTime'].append(img.isot)
                sos_files = []
                # img_mjd = int(Time(img.isot).mjd)
                # All boss exposures write as splog, but manga writes different
                red_dir = sdss_paths.sos / f"{self.args.sjd}"
                red_fil = red_dir / 'splog-r1-{:0>8}.log'.format(img.exp_id)
                if red_fil.exists():
                    sos_files.append('r1')
                else:
                    sos_files.append('xx')
                if (red_fil.parent / red_fil.name.replace('r1', 'b1')).exists():
                    sos_files.append('b1')
                else:
                    sos_files.append('xx')
                if (red_fil.parent / red_fil.name.replace('r1', 'r2')).exists():
                    sos_files.append('r2')
                # else:
                #     m_detectors.append('xx')
                if (red_fil.parent / red_fil.name.replace('r1', 'b2')).exists():
                    sos_files.append('b2')
                # else:
                #     m_detectors.append('xx')
                self.b_data['iDetector'].append('-'.join(sos_files))

    def sort(self):
        """Sorts self.ap_data by design time and by image time and converts to
        arrays"""
        # Data
        for key, item in self.data.items():
            if 'Time' in key:
                try:
                    self.data[key] = Time(item)
                except ValueError:
                    self.data[key] = Time(item, format='isot')
            else:
                self.data[key] = np.array(item)
        data_sort = self.data['dTime'].argsort()
        for key, item in self.data.items():
            self.data[key] = item[data_sort]

        if self.args.apogee:
            for key, item in self.ap_data.items():
                if 'Time' in key:
                    try:
                        self.ap_data[key] = Time(item)
                    except ValueError:
                        self.ap_data[key] = Time(item, format='isot')
                else:
                    self.ap_data[key] = np.array(item)
            ap_design_sorter = self.ap_data['dTime'].argsort()
            ap_img_sorter = self.ap_data['iTime'].argsort()
            ap_dome_sorter = self.ap_data['fTime'].argsort()
            ap_arc_sorter = self.ap_data['aTime'].argsort()
            ap_obj_sorter = self.ap_data['oTime'].argsort()
            for key, item in self.ap_data.items():
                if key[0] == 'd':
                    self.ap_data[key] = item[ap_design_sorter]
                elif key[0] == 'i':
                    self.ap_data[key] = item[ap_img_sorter]
                elif key[0] == 'f':
                    self.ap_data[key] = item[ap_dome_sorter]
                elif key[0] == 'a':
                    self.ap_data[key] = item[ap_arc_sorter]
                elif key[0] == 'o':
                    self.ap_data[key] = item[ap_obj_sorter]
            if self.args.morning:
                was_dark = False
                prev_time = 0
                lower = None
                for t, reads, exp in zip(self.ap_data['iTime'],
                                         self.ap_data['iNRead'],
                                         self.ap_data['iEType']):
                    if (reads == 60) and (exp == 'Dark'):
                        if was_dark:
                            lower = prev_time
                            break
                        else:
                            was_dark = True
                            prev_time = t
                            if self.args.verbose:
                                print('Morning lower limit: {}'.format(
                                    prev_time))
                    else:
                        was_dark = False
                upper = Time(self.args.sjd + 1, format='mjd')
                if lower is None:
                    raise Exception('Morning cals not completed for this date')
                self.morning_filter = ((lower <= self.ap_data['iTime'])
                                       & (self.ap_data['iTime'] <= upper))

        if self.args.boss:
            for key, item in self.b_data.items():
                if 'Time' in key:
                    try:
                        self.b_data[key] = Time(item)
                    except ValueError:
                        self.b_data[key] = Time(item, format='isot')
                else:
                    self.b_data[key] = np.array(item)
            b_design_sorter = self.b_data['dTime'].argsort()
            b_img_sorter = self.b_data['iTime'].argsort()
            b_h_sorter = self.b_data['hTime'].argsort()
            for key, item in self.b_data.items():
                if key[0] == 'c':
                    self.b_data[key] = item[b_design_sorter]
                elif key[0] == 'i':
                    self.b_data[key] = item[b_img_sorter]
                elif key[0] == 'h':
                    self.b_data[key] = item[b_h_sorter]

    def count_dithers(self):
        for i, design in enumerate(self.data['dDesign']):
            self.design_data['cNAPA'].append(np.sum(
                (self.ap_data['iDesign'] == design)
                & (self.ap_data['iDither'] == 'A')
                & (self.ap_data['iEType'] == 'Object')))
            self.design_data['cNAPB'].append(np.sum(
                (self.ap_data['iDesign'] == design)
                & (self.ap_data['iDither'] == 'B')
                & (self.ap_data['iEType'] == 'Object')))
            # self.design_data['cNBN'].append(np.sum(
            # (self.b_data['iDesign'] == design)
            # & (self.b_data['iDither'] == 'N')
            # & (self.b_data['iEType'] == 'Science')))
            # self.design_data['cNBS'].append(np.sum(
            # (self.b_data['iDesign'] == design)
            # & (self.b_data['iDither'] == 'S')
            # & (self.b_data['iEType'] == 'Science')))
            # self.design_data['cNBE'].append(np.sum(
            # (self.b_data['iDesign'] == design)
            # & (self.b_data['iDither'] == 'E')
            # & (self.b_data['iEType'] == 'Science')))
            # self.design_data['cNBC'].append(np.sum(
            # (self.b_data['iDesign'] == design)
            # & (self.b_data['iDither'] == 'C')
            # & (self.b_data['iEType'] == 'Science')))
            self.design_data['cNB'].append(np.sum(
                (self.b_data['iDesign'] == design)
                & (self.b_data['iEType'] == 'Science')))
            if self.design_data['cNB'][-1] != 0:
                self.design_data['cBdt'].append(np.max(
                    self.b_data['idt'][
                        (self.b_data['iDesign'] == design)
                        & (self.b_data['iEType'] == 'Science')]))
            else:
                self.design_data['cBdt'].append(0)

        for i, design in enumerate(self.data['dDesign']):
            """To determine the number of apogee a dithers per design (cNAPA),
            as well as b dithers (cNAPB), and the same for NSE dithers."""
            # APOGEE dithers
            if self.design_data['cNAPA'][i] == self.design_data['cNAPB'][i]:
                if self.design_data['cNAPA'][i] != 0:
                    self.design_data['cAPSummary'].append(
                        '{}xAB'.format(self.design_data['cNAPA'][i]))
                else:
                    self.design_data['cAPSummary'].append('No APOGEE Science')
            else:
                self.design_data['cAPSummary'].append(
                    '{}xA {}xB'.format(self.design_data['cNAPA'][i],
                                       self.design_data['cNAPB'][i]))
            # BOSS (MaNGA) dithers
            # if self.design_data['cNBC'][i] == 0:
            # if (self.design_data['cNBN'][i]
            # == self.design_data['cNBS'][i]
            # == self.design_data['cNBE'][i]):
            # self.design_data['cBSummary'].append(
            # '{}xNSE'.format(self.design_data['cNBN'][i]))
            # else:
            # self.design_data['cBSummary'].append(
            # '{}xN {}xS {}xE'.format(self.design_data['cNBN'][i],
            # self.design_data['cNBS'][i],
            # self.design_data['cNBE'][i]))
            # else:
            if self.design_data['cNB'][i] != 0:
                self.design_data['cBSummary'].append(
                    '{}x{}s'.format(self.design_data['cNB'][i],
                                    self.design_data['cBdt'][i]))
            else:
                self.design_data['cBSummary'].append('No BOSS Science')

    @staticmethod
    def hartmann_parse(hart):
        output = '{}\n'.format(str(hart[0].times[-1])[:19])
        output += 'r1: {:>6.0f}, b1: {:>6.1f}\n'.format(
            hart[0].values[-1], hart[1].values[-1])
        # output += 'r2: {:>6.0f}, b2: {:>6.1f}\n'.format(
        #  hart[1].values[-1], hart[3].values[-1])
        output += 'Average Move: {:>6.0f}\n'.format(hart[2].values[-1])
        # output += 'SP1: {:>6.0f}, SP2: {:>6.0f}\n'.format(
        # hart[4].values[-1], hart[5].values[-1])
        output += 'R Residuals: {:>6.0f}\n'.format(hart[3].values[-1])
        # output += 'SP1: {:>6.0f}, SP2: {:>6.0f}\n'.format(
        # hart[10].values[-1], hart[11].values[-1])
        output += 'B Residuals: {:>6.1f}\n'.format(hart[4].values[-1])
        # output += 'SP1: {:>6.1f}, SP2: {:>6.1f}\n'.format(
        # hart[6].values[-1], hart[7].values[-1])
        output += 'SP1 Temperature: {:>6.1f}\n'.format(hart[5].values[-1])
        # output += 'SP1: {:>6.1f}, SP2: {:>6.1f}'.format(
        # hart[8].values[-1], hart[9].values[-1])
        return output

    def p_summary(self):
        print('=' * 80)
        print('{:^80}'.format('Observing Summary'))
        print('=' * 80)
        for i, design  in enumerate(self.data['dDesign']):
            print('')
            print("Design {}, {}, {}".format(design, self.data['dConfig'][i],
                                 self.design_data['cAPSummary'][i],
                                 self.design_data['cBSummary'][i]))
        print()
        if len(self.ap_data["fRatio"]) > 0:
            flux_ratio = np.nanmean(np.array(self.ap_data["fRatio"]), axis=0)
            avg = np.nanmean(flux_ratio)
            missing = flux_ratio < 0.2
            faint = (flux_ratio < 0.7) & (0.2 <= flux_ratio)
            bright = ~missing & ~faint
            i_missing = np.where(missing)[0].astype(int) + 1
            i_faint = np.where(faint)[0].astype(int) + 1
            i_bright = np.where(bright)[0]
            if self.ap_image is not None:
                missing_bundles = self.ap_image.create_bundles(i_missing)
                faint_bundles = self.ap_image.create_bundles(i_faint)
                print("APOGEE Dome Flats\n"
                    f"Missing Fibers: {missing_bundles}\n"
                    f" Faint fibers: {faint_bundles}\n"
                    f" Average Throughput: {avg:.3f}")
                print()

        print('### Notes:\n')
        start_time = Time(self.args.sjd, format="mjd")
        end_time = Time(self.args.sjd + 1, format="mjd")
        dust_sum = get_dust.get_dust(start_time, end_time, self.args.verbose)
        print('- Integrated Dust Counts: ~{:5.0f} dust-hrs'.format(
            dust_sum - dust_sum % 100))
        print('\n')

        # print('=' * 80)
        # print('{:^80}'.format('Comments Timeline'))
        # print('=' * 80)
        # print()

    @staticmethod
    def get_window(data, i):
        try:
            window = ((data['iTime']
                       >= data['dTime'][i])
                      & (data['iTime']
                         < data['dTime'][i + 1])
                      )

        except IndexError:
            try:
                window = ((data['iTime'] >= data['dTime'][i])
                          & (data['iTime'] < Time.now() + 0.3))
            except IndexError:
                window = np.array([False] * len(data['iTime']))

        return window

    def p_data(self):
        print('=' * 80)
        print('{:^80}'.format('Data Log'))
        print('=' * 80 + '\n')
        for i, design in enumerate(self.data['dDesign']):
            print('### Design {}\n'.format(design))
            if design  in self.ap_data['dDesign']:
                ap_design = np.where(design == self.ap_data['dDesign'])[0][0]

                print('# APOGEE')
                print('{:<5} {:<8} {:<8} {:<12} {:<4} {:<6} {:<5}'
                      ' {:<4}'.format('MJD', 'UTC', 'Exposure', 'Type',
                                      'Dith', 'Reads', 'Arch',
                                      'Seeing'))
                print('-' * 80)
                window = self.get_window(self.ap_data, ap_design)
                for (mjd, iso, exp_id, exp_type, dith, nread,
                     detectors, see) in zip(
                    self.ap_data['iTime'][window].mjd + 0.3,
                    self.ap_data['iTime'][window].iso,
                    self.ap_data['iID'][window],
                    self.ap_data['iEType'][window],
                    self.ap_data['iDither'][window],
                    self.ap_data['iNRead'][window],
                    self.ap_data['iDetector'][window],
                    self.ap_data['iSeeing'][window]
                ):
                    print('{:<5.0f} {:0>8} {:<8.0f} {:<12} {:<4} {:>5} {:<5}'
                          ' {:>4.1f}'.format(int(mjd), iso[11:19], exp_id,
                                             exp_type,
                                             dith, nread, detectors, see))
                print()
                if design in self.ap_data['dDesign']:
                    for j, dome in enumerate(self.ap_data['fDesign']):
                        if dome == design:
                            print(self.ap_data['fTime'][j].iso)
                            print(textwrap.fill('Missing fibers: {}'.format(
                                self.ap_data['fMissing'][j]), 80))
                            print(textwrap.fill('Faint fibers: {}'.format(
                                self.ap_data['fFaint'][j]), 80))
                            print()

            if design in self.b_data['dDesign']:
                print('# BOSS')
                print('{:<5} {:<8} {:<8} {:<7} {:<4} {:<11} {:<5} {:<5}'
                      ''.format('MJD', 'UTC', 'Exposure', 'Type',
                                'Dith', 'SOS', 'ETime', 'Hart'))
                print('-' * 80)
                # i is an index for data, but it will disagree with b_data
                # if there is an apogee-onlydesign 
                b_design = np.where(design == self.b_data['dDesign'])[0][0]
                window = self.get_window(self.b_data, b_design)
                for (mjd, iso, exp_id, exp_type, dith,
                     detectors, etime, hart) in zip(
                    self.b_data['iTime'][window].mjd + 0.3,
                    self.b_data['iTime'][window].iso,
                    self.b_data['iID'][window],
                    self.b_data['iEType'][window],
                    self.b_data['iDither'][window],
                    self.b_data['iDetector'][window],
                    self.b_data['idt'][window],
                    self.b_data['iHart'][window],
                ):
                    print('{:<5.0f} {:0>8} {:0>8.0f} {:<7} {:<4} {:<11}'
                          ' {:>5.0f} {:<5}'
                          ''.format(int(mjd), iso[11:19], exp_id,
                                    exp_type.strip(),
                                    dith.strip(), detectors, etime,
                                    hart))
                try:
                    window = ((self.b_data['hTime']
                               >= self.data['dTime'][i])
                              & (self.b_data['hTime']
                                 < self.data['dTime'][i + 1])
                              )

                except IndexError:
                    window = ((self.b_data['hTime']
                               >= self.data['dTime'][i])
                              & (self.b_data['hTime'] < Time.now()))
                if self.b_data['hTime'][window]:
                    print()
                    # print('Hartmanns')
                for t, hart in zip(self.b_data['hTime'][window],
                                   self.b_data['hHart'][window]):
                    print(self.hartmann_parse(hart))
                print()

    def p_boss(self):
        print('=' * 80)
        print('{:^80}'.format('BOSS Data Summary'))
        print('=' * 80 + '\n')
        print('{:<5} {:<8} {:<13} {:<8} {:<7} {:<4} {:<11} {:<5} {:<5}'
              ''.format('MJD', 'UTC', 'Design-Config', 'Exposure', 'Type', 'Dith',
                        'SOS', 'ETime', 'Hart'))
        print('-' * 80)
        for (mjd, iso, design, design, config, exp_id, exp_type, dith, detectors, etime,
             hart) in zip(self.b_data['iTime'].mjd + 0.3,
                          self.b_data['iTime'].iso,
                          self.b_data['iDesign'],
                          self.b_data["iDesign"],
                          self.b_data['iConfig'],
                          self.b_data['iID'],
                          self.b_data['iEType'],
                          self.b_data['iDither'],
                          self.b_data['iDetector'],
                          self.b_data['idt'],
                          self.b_data['iHart']):
            print('{:<5.0f} {:>8} {:>6}-{:<6.0f} {:0>8.0f} {:<7} {:<4}'
                  ' {:<11}'
                  ' {:>5.0f} {:<5}'
                  ''.format(int(mjd), iso[11:19], design, config, exp_id,
                            exp_type.strip(),
                            dith.strip(), detectors, etime, hart))
        print()

    def p_apogee(self):
        print('=' * 80)
        print('{:^80}'.format('APOGEE Data Summary'))
        print('=' * 80 + '\n')
        print('{:<5} {:<8} {:<13} {:<8} {:<12} {:<4} {:<5} {:<5}'
              ''.format('MJD', 'UTC', 'Design-Config', 'Exposure', 'Type',
                              'Dith', 'Reads', 'Arch'))
        print('-' * 80)
        if self.args.morning:
            for (mjd, iso, design, config, exp_id, exp_type, dith, nread,
                 detectors) in zip(
                self.ap_data['iTime'].mjd[self.morning_filter] + 0.3,
                self.ap_data['iTime'].iso[self.morning_filter],
                self.ap_data["iDesign"][self.morning_filter],
                self.ap_data["iConfig"][self.morning_filter],
                self.ap_data['iID'][self.morning_filter],
                self.ap_data['iEType'][self.morning_filter],
                self.ap_data['iDither'][self.morning_filter],
                self.ap_data['iNRead'][self.morning_filter],
                self.ap_data['iDetector'][self.morning_filter],
            ):
                print('{:<5.0f} {:>8} {:>6.0f}-{:<6.0f} {:<8.0f} {:<12} {:<4}'
                      ' {:>5}'
                      ' {:<5}'
                      ''.format(int(mjd), iso[11:19], design, config,
                                         exp_id, exp_type,
                                         dith, nread, detectors))

        else:
            for (mjd, iso, design, config, exp_id, exp_type, dith, nread,
                 detectors) in zip(
                self.ap_data['iTime'].mjd,
                self.ap_data['iTime'].iso,
                self.ap_data["iDesign"],
                self.ap_data["iConfig"],
                self.ap_data['iID'], self.ap_data['iEType'],
                self.ap_data['iDither'], self.ap_data['iNRead'],
                self.ap_data['iDetector'],
            ):
                # print('{:<5.0f} {:>8} {:>2.0f}-{:<5.0f} {:<8.0f} {:<12} {:<4}'
                #       ' {:>6}'
                #       ' {:<8}'
                #       ' {:>6.1f}'.format(int(mjd), iso[11:19], design, plate,
                #                          exp_id, exp_type,
                #                          dith, nread, detectors, see))
                print('{:<5.0f} {:>8} {:>6.0f}-{:<6.0f} {:<8.0f} {:<12} {:<4}'
                      ' {:>5}'
                      ' {:<5}'
                      ''.format(int(mjd), iso[11:19], design, config,
                                         exp_id, exp_type,
                                         dith, nread, detectors))

        # Usually, there are 4 ThAr and 4 UNe arcs in a night, and they're
        # assumed to be alternating ThAr UNe ThAr UNe. When you grab every
        # other, you'll have only one type, that's the first slicing, and the
        # second slicing is that you only care about the diffs between two
        # dithers taken back to back.
        wrapper = textwrap.TextWrapper(80)
        thar_str = 'ThAr Offsets: {}'.format(
            ['{:.2f}'.format(f) for f in np.diff(
                self.ap_data['aOffset'][self.ap_data['aLamp'] == 'ThAr'])])
        print('\n'.join(wrapper.wrap(thar_str)))
        une_str = 'UNe Offsets: {}'.format(
            ['{:.2f}'.format(f) for f in np.diff(
                self.ap_data['aOffset'][self.ap_data['aLamp'] == 'UNe'])])
        print('\n'.join(wrapper.wrap(une_str)))
        if len(self.ap_data['oOffset']) > 1:
            # Put it under an if in case we didn't open.
            rel_offsets = np.abs(np.diff(self.ap_data['oOffset']))
            obj_str = ('Object Offsets: Max: {:.2f}, Min: {:.2f}, Mean: {:.2f}'
                       ''.format(rel_offsets.max(), rel_offsets.min(),
                                 rel_offsets.mean()))
            # ['{:>6.3f}'.format(f) for f in np.diff(
            #     self.ap_data['oOffset'])])
            print('\n'.join(wrapper.wrap(obj_str)))
        # obj_offsets = []
        # prev_dither = None
        # prev_f = 0.
        # for d, f in zip(self.ap_data['oDither'], self.ap_data['oOffset']):
        #     if d != prev_dither:
        #         obj_offsets.append('{:.3f}'.format(f - prev_f))
        #     prev_dither = d
        #     prev_f = f
        # obj_str = 'Object Offsets: {}'.format(obj_offsets)
        print('\n')

    def log_support(self):
        print('=' * 80)
        print(f"{'Log Support':^80}")
        print('=' * 80)
        start = Time(self.args.sjd - 0.3, format='mjd')
        end = Time(self.args.sjd + 1, format='mjd') - 0.3
        end = Time.now() if Time.now() < end else end
        tel = log_support.LogSupport(start, end, self.args)
        tel.set_callbacks()
        tel.get_offsets()
        tel.get_focus()
        tel.get_weather()
        tel.get_hartmann()
        print(tel.offsets)
        print(tel.focus)
        print(tel.weather)
        print(tel.hartmann)

    @staticmethod
    def mirror_numbers():
        print('=' * 80)
        print('{:^80}'.format('Mirror Numbers'))
        print('=' * 80 + '\n')
        try:
            mirror_nums = m4l.mirrors()
            print(mirror_nums)
        except (ConnectionRefusedError, TimeoutError) as me:
            print('Could not fetch mirror numbers:\n{}'.format(me))

    @staticmethod
    def tel_status():
        print('=' * 80)
        print('{:^80}'.format('Telescope Status'))
        print('=' * 80 + '\n')
        try:
            status = telescope_status.query()
            print(status)
        except OSError as oe:
            print(f"Couldn't get telescope status:\n{oe}")


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--today', action='store_true', default=True,
                        help="Whether or not you want to search for today's"
                             " data, whether or not the night is complete."
                             " Note: must be run after 18:00Z to get the"
                             " correct sjd")
    parser.add_argument('-m', '--mjd', type=int,
                        help='If not today (-t), the mjd to search (actually'
                             ' sjd)')
    parser.add_argument('--mirrors', '--mirror', action='store_true',
                        help='Print mirror numbers using m4l.py')
    parser.add_argument("-s", '--summary', help='Print the data summary',
                        action='store_true')
    parser.add_argument('-d', '--data', action='store_true',
                        help='Print the data log')
    parser.add_argument('-p', '--print', action='store_true',
                        help='Print all possible outputs')
    parser.add_argument('-b', '--boss', action='store_true',
                        help='Print BOSS Summary')
    parser.add_argument('-a', '--apogee', action='store_true',
                        help='Print APOGEE Summary')
    parser.add_argument('-l', '--log-support', action='store_true',
                        help='Print 4 log support sections')
    parser.add_argument('-n', '--noprogress', action='store_true',
                        help='Show no progress in processing images. WARNING:'
                             ' Might be slower, but it could go either way.')
    parser.add_argument("-t", '--telstatus', action='store_true',
                        help='Print telescope status')
    parser.add_argument('--morning', action='store_true',
                        help='Only output apogee morning cals')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Increased printing for debugging')
    parser.add_argument('--legacy-aptest', action='store_true',
                        help='Use utr_cdr images for aptest instead of'
                             ' quickred for the ap_test')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    if args.mjd:
        args.sjd = args.mjd
    elif args.today:
        args.sjd = sjd.sjd()
    else:
        raise argparse.ArgumentError(args.sjd,
                                     'Must provide -t or -m in arguments')
    if args.verbose:
        print(args.sjd)
    ap_data_dir = ap_dir / '{}'.format(args.sjd)
    b_data_dir = b_dir / '{}'.format(args.sjd)
    ap_images = Path(ap_data_dir).glob('apR-a*.apz')
    b_images = Path(b_data_dir).glob('sdR-r1*fit.gz')

    if not args.noprogress:
        try:
            ap_images = list(ap_images)
            b_images = list(b_images)
        except OSError:  # Stale NFS handle
            ap_images = list(ap_images)
            b_images = list(b_images)
    p_boss = args.boss
    p_apogee = args.apogee

    if args.print:
        args.summary = True
        args.data = True
        args.boss = True
        args.apogee = True
        p_boss = True
        p_apogee = True
        # args.log_support = True  # Because this usually produces wrong results
        args.mirrors = True
        args.telstatus = True

    if args.summary:
        args.boss = True
        args.apogee = True

    if args.morning:
        args.apogee = True
        p_apogee = True

    if args.data:
        args.boss = True
        args.apogee = True

    log = Logging(ap_images, b_images, args)
    log.parse_images()
    log.sort()
    log.count_dithers()

    if args.summary:
        log.p_summary()

    if args.data:
        log.p_data()

    if p_apogee:
        log.p_apogee()

    if p_boss:
        log.p_boss()

    if args.log_support:
        log.log_support()

    if args.mirrors:
        pass
        # log.mirror_numbers()

    if args.telstatus:
        log.tel_status()
    return log


if __name__ == '__main__':
    main()
